# coding: utf8
__doc__ = 'subprocess task control'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from time import sleep
try:
    from Queue import Empty
except ImportError:
    from queue import Empty
from threading import Thread
from multiprocessing import Process, Queue
try:
    from termios import error as termios_error
except ImportError:
    termios_error = None
from . import Env
from . import Instrument
from os.path import join, dirname
import gettext
_t = gettext.translation( "mcush", join(dirname(__file__), "locale"), codeset='utf8', fallback=True )
try:
    _getStr = _t.ugettext
except AttributeError:
    _getStr = _t.gettext


class Task():
    STOP_MAX_DELAY = 15
    STOP_CHECK_PERIOD = 0.2

    def __init__( self, args, queue ):
        self.started = False
        self.args = args
        self.queue = queue
        self.control_queue = Queue()
        self.p = Process( target=self.monitor )
        self.p.start()
        self.listener = Thread( target=self.listener )
        self.listener.setDaemon( 1 )
        self.listener.start()
        self.manual_stop = False
        self.started = True

    def target( self, args ):
        raise NotImplementedError

    def waitStop( self, timeout ):
        try:
            ret = self.control_queue.get( block=True, timeout=timeout  )
            assert ret == 'stop', ret
            self.info( _getStr('Stopping') )
            return True
        except Empty:
            pass 
        return False

    def stop( self ):
        if not self.p.is_alive():
            return True
        self.manual_stop = True
        self.control_queue.put( 'stop' )
        for i in range( int(self.STOP_MAX_DELAY / self.STOP_CHECK_PERIOD) ):
            if not self.p.is_alive():
                return True
            else:
                sleep( self.STOP_CHECK_PERIOD )
        if Env.VERBOSE:
            print( 'subprocess still alive' )
        return False

    def info( self, info, info_type='info' ):
        self.queue.put( ('info', (info, info_type)) )

    def end( self ):
        self.queue.put( ('lock', False) )
        self.queue.put( ('end', False) )

    def monitor( self ):
        try:
            self.queue.put( ('lock', True) )
            self.target( self.args )
        except Instrument.CommandTimeoutError:
            self.info( _getStr('Stopping') )
            self.info( _getStr('Timeout'), 'error' )
        except Instrument.PortNotFound as e:
            self.info( _getStr('Port not exist') + ': ' + unicode(e), 'error' )
        except Instrument.UnknownPortError as e:
            self.info( _getStr('Unknown port error') + ': ' + unicode(e), 'error' )
        except Instrument.IDNMatchError as e:
            self.info( _getStr('IDN not match') + ': ' + unicode(e), 'error' )
        except Exception as e:
            if Env.VERBOSE:
                print( type(e), str(e) )
            info = _getStr('Unknown error') + ': '
            if termios_error and isinstance(e, termios_error):
                a,b = e.args
                info += u'%d %s'% (a, b.decode(encoding='utf8'))
            else:
                #for c in str(e):  # e string may not be Unicode
                #    info += c if ord(c) <= 128 else 'x%X'% ord(c)
                info = str(e).decode(encoding='utf8')
            self.info( info, 'error' )
     
    def listener( self ):
        if self.p.is_alive():
            #print 'listener:alive'
            self.p.join()
            #print 'listener:end'
        self.end()
       
    def isStopped( self ):
        return not self.p.is_alive()

    def isRunning( self ):
        return self.p.is_alive()
  
    def terminate( self, safely=False ): 
        if safely:
            if self.stop():
                return
        if self.p.is_alive():
            self.p.terminate()
        self.info( _getStr('Stopped') )
        self.end()

    def pause( self, prompt ):
        # clear queue
        while True:
            try:
                self.control_queue.get( block=False )
            except Empty:
                break
        self.queue.put( ('pause', prompt) )
        try:
            ret = self.control_queue.get( block=True, timeout=None )
            if ret == 'continue':
                return True 
        except Empty:
            pass
        return False

    def halt( self ):
        self.queue.put( ('halt', None) )
 
    def done( self ):
        self.queue.put( ('done', None) )





# coding: utf8
__doc__ = '3rd party modules, addons'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'

from . import BeepPlayer
from . import MorseCodeBeeper
from . import Tjc
from . import Canvas
from . import Gratten
from . import SegmentTable
from . import Tm1637
from . import Bitmap
from . import Font
from . import Max7219
from . import Ssd1306
from . import Ws2812
from . import Motion 
from . import Dallas1W
from . import Aosong
from . import Max6675
from . import Max31865
from . import Tm7707
from . import Cs1238
from . import Hx711
from . import Ina219
from . import Bosch
from . import Feeltech
from . import Tcs3472
from . import Max30102
from . import Ads1115
from . import Hmc5883
from . import Korad
from . import Yaohua
from . import Tongmen




# coding: utf8
__doc__ = 'TCS3472 color sensor from ams AG'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'


REG_ENABLE = 0x00
REG_ATIME = 0x01
REG_WTIME = 0x03
REG_AILT = 0x04
REG_AIHT = 0x06
REG_PERS = 0x0C
REG_CONFIG = 0x0D
REG_CONTROL = 0x0F
REG_ID = 0x12
REG_STATUS = 0x13
REG_CDATA = 0x14
REG_RDATA = 0x16
REG_GDATA = 0x18
REG_BDATA = 0x1A


class Tcs3472():
        
    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x29, scl, sda )
        self.init()

    def readReg( self, addr ):
        command = [ 0x80 | (1<<5) | (addr&0x1F) ]
        return self.controller.i2c( command, read=1 )[0]

    def readReg2( self, addr ):
        command = [ 0x80 | (1<<5) | (addr&0x1F) ]
        r = self.controller.i2c( command, read=2 )
        return (r[1]<<8)+r[0]

    def writeReg( self, addr, dat ):
        command = [ 0x80 | (1<<5) | (addr&0x1F), dat&0xFF]
        self.controller.i2c( command )

    def writeReg2( self, addr, dat ):
        command = [ 0x80 | (1<<5) | (addr&0x1F), dat&0xFF, (dat>>8)&0xFF]
        self.controller.i2c( command )

    def init( self ):
        self.id = self.readReg( REG_ID )  # chip id
        self.writeReg( REG_ENABLE, 0x03 )  # power on, adc enabled
        self.writeReg( REG_ATIME, 0x00 )  # integ_cycles:256, time 700ms
        self.writeReg( REG_WTIME, 0x00 )  # wait:256, time 614ms

    def readStatus( self ):
        return self.readReg( REG_STATUS )

    def readRawData( self ):
        c = self.readReg2( REG_CDATA )
        r = self.readReg2( REG_RDATA )
        g = self.readReg2( REG_GDATA )
        b = self.readReg2( REG_BDATA )
        return c, r, g, b




# coding: utf8
__doc__ = 'tm1637 led controller module'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush
from . import SegmentTable


BRIGHTNESS = 7

CMD_DATA = 0x40
CMD_DATA_READ_KEY = 0x42
CMD_DATA_FIXED_ADDR = 0x44
CMD_DATA_TEST_MODE = 0x48
CMD_CONTROL = 0x80
CMD_ADDRESS = 0xC0


class LED4:
    DIGITS = 4
    
    def __init__( self, controller, scl=None, sda=None, brightness=BRIGHTNESS ):
        self.controller = controller
        self.segments = [0]*self.DIGITS
        self.controller.i2c_init( 0, scl=scl, sda=sda, lsb=True )
        self.brightness = brightness
        
    def clear( self ):
        self.display_segs( 0, "    ", ascii_mode=True )

    def display_string( self, string, index=0 ):
        segments = []
        for c in string:
            if not c in ['.', ':']:
                segments.append( SegmentTable.SEGMENT_TABLE_ASCII[ord(c)] )
            elif len(segments) > 0:
                segments[-1] |= SegmentTable.SH
        self.controller.i2c( addr=CMD_DATA )
        self.controller.i2c( segments[:self.DIGITS], addr=CMD_ADDRESS+index )
        self.controller.i2c( addr=CMD_CONTROL+0x08+self.brightness ) 

    def display_integer( self, val, hex_mode=False ):
        if hex_mode:
            string = (('%%%dX'% self.DIGITS)% val)[-self.DIGITS:]
        else:
            string = (('%%%dd'% self.DIGITS)% val)[-self.DIGITS:]
        self.display_segs( 0, string[:self.DIGITS], ascii_mode=True )
    
    def display_segment( self, index, segment, ascii_mode=None, hex_mode=None ):
        self.controller.i2c( addr=CMD_DATA+index )
        if hex_mode:
            segment = SegmentTable.SEGMENT_TABLE_HEX[segment]
        elif ascii_mode:
            segment = SegmentTable.SEGMENT_TABLE_ASCII[ord(segment)]
        self.segments[index] = segment
        self.controller.i2c( [segment], addr=CMD_ADDRESS+index )
        self.controller.i2c( addr=CMD_CONTROL+0x08+self.brightness ) 

    def display_segments( self, index, segments, ascii_mode=None, hex_mode=None ):
        self.controller.i2c( addr=CMD_DATA )
        segments = segments[:self.DIGITS]
        if hex_mode:
            segments = [ SegmentTable.SEGMENT_TABLE_HEX[c] for c in segments ]
        elif ascii_mode:
            segments = [ SegmentTable.SEGMENT_TABLE_ASCII[ord(c)] for c in segments ]
        self.segments = segments
        self.controller.i2c( segments, addr=CMD_ADDRESS+index )
        self.controller.i2c( addr=CMD_CONTROL+0x08+self.brightness ) 

    clrscr = clear
    display_str = display_string
    display_int = display_integer
    display_seg = display_segment
    display_segs = display_segments


class LED8:
    DIGITS = 8





# coding:utf8
__doc__ = 'DP800 Multi-channels Programmable Power'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env

# Product lines:
#        Channel  Power  Output(per channel)
# DP811  1        200W   50V/3A    
# DP821  2        140W   8V10A||60V/1A    
# DP831  3        160W   8V5A||30V/2A,-30V/2A
# DP832  3        195W   30V/3A||30V/3A,5V/3A
#
class DP800( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'DP800'
    DEFAULT_IDN = re_compile( '.*' )

    def __init__( self, *args, **kwargs ):
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 
        raise NotImplementedError

        # TODO: append the instruments




# coding: utf8
__doc__ = 'MAX6675 SPI Thermometer'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils

class Max6675():
        
    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None ):
        self.controller = controller
        self.controller.spi_init( sdi, sdo, sck, cs, width=16 )
 
    def getTemperature( self ):
        v = self.controller.spiRead([0])[0]
        t = ((v>>3)&0xFFF)/4.0  # in degree
        return t




# coding: utf8
__doc__ = 'Honeywell HMC5583 3-Axis Compass'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Instrument, Utils
import time


REG_CONFIG_A  = 0
REG_CONFIG_B  = 1
REG_MODE      = 2
REG_DAT_X_H   = 3
REG_DAT_X_L   = 4
REG_DAT_Z_H   = 5
REG_DAT_Z_L   = 6
REG_DAT_Y_H   = 7
REG_DAT_Y_L   = 8
REG_STATUS    = 9
REG_ID_A      = 10
REG_ID_B      = 11
REG_ID_C      = 12

LSB_PER_GAUSS_GAIN = [1370, 1090, 820, 660, 440, 390, 330, 230]

class Hmc5883():

    def __init__( self, controller, scl=None, sda=None, addr_mode=0 ):
        self.controller = controller
        self.controller.i2c_init( 0x1E, scl, sda )
        self.config()
        self.mode()
        
    def config( self, ma=3, do=4, ms=0, gain=1 ):
        a = (ma<<5)+(do<<2)+(ms)
        b = (gain<<5)
        self.controller.i2c( [REG_CONFIG_A, a, b] )
        self.gain = gain

    def mode( self, continuous=True ):
        m = 0 if continuous else 1
        self.controller.i2c( [REG_MODE, m] )

    def readID( self ):
        a, b, c = self.controller.i2c( [REG_ID_A], read=3 )
        return (a<<16)+(b<<8)+c

    def readStatus( self ):
        s = self.controller.i2c( [REG_STATUS], read=1 )[0]
        return s

    def readValue( self ):
        while True:
            if self.readStatus() & 0x01:
                break
            time.sleep(0.01)
        xh, xl, zh, zl, yh, yl = self.controller.i2c( [REG_DAT_X_H], read=6 )
        x = Utils.H2h((xh<<8)+xl)
        z = Utils.H2h((zh<<8)+zl)
        y = Utils.H2h((yh<<8)+yl)
        lpg = float(LSB_PER_GAUSS_GAIN[self.gain])
        return (x/lpg, y/lpg, z/lpg)




# coding: utf8
__doc__ = 'Products from FLUKE Corporation.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'

from . import F8808A
from . import F5520A




# coding:utf8
__doc__ = 'kivy on Android'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
import os
import re
import sys
import time
import base64
import logging
from .. import Env
from .. import Utils
from .. import Instrument
from .. import Mcush



class JavaBluetoothPort(Instrument.Port):
        
    def __init__( self, parent, *args, **kwargs ):
        Instrument.Port.__init__( self, parent, *args, **kwargs )
        from jnius import autoclass
        self.UUID = autoclass('java.util.UUID')
        self.BluetoothAdapter = autoclass('android.bluetooth.BluetoothAdapter')
        self.BluetoothDevice = autoclass('android.bluetooth.BluetoothDevice')
        self.BluetoothSocket = autoclass('android.bluetooth.BluetoothSocket')
        
    def connect( self ):
        if self._connected:
            return
        paired_devices = self.BluetoothAdapter.getDefaultAdapter().getBondedDevices().toArray()
        for device in paired_devices:
            if device.getName() == self.port:
                self.socket = device.createRfcommSocketToServiceRecord(self.UUID.fromString("00001101-0000-1000-8000-00805F9B34FB"))
                #print( type(self.socket), self.socket )
                self.recv = self.socket.getInputStream()
                self.send = self.socket.getOutputStream()
                print( 'JavaBluetoothPort.connect', 'recv', type(self.recv), self.recv, 'send',  type(self.send), self.send )
                self.socket.connect()
                self._connected = True
                break

    def disconnect( self ):
        self._connected = False
        self.socket.close()
    
    def read( self, read_bytes=1 ):
        if self._connected:
            ret = []
            for i in range(read_bytes):
                r = self.recv.read()
                if r:
                    ret.append( chr(r) )
                else:
                    break
            #print( 'read', ret )
            return ''.join(ret)

    def write( self, buf ):
        if self._connected:
            self.send.write(buf)
 
    def flush( self ):
        if self._connected:
            self.send.flush()


class KMcush( Mcush.Mcush ):
    '''Mcush on Android/Kivy'''
    PORT_TYPE = JavaBluetoothPort

    def writeCommand( self, cmd ):
        try:
            ret = Mcush.Mcush.writeCommand( self, cmd )
            print( 'cmd: ' + cmd )
            print( 'ret: ' + '\n'.join(ret) )
            return ret
        except Instrument.CommandSyntaxError:
            print( 'command syntax err: ' + cmd )
            return []

    def vibrate( self, duration=0.05 ):
        try:
            self.vibrator
        except AttributeError:
            from plyer import vibrator
            self.vibrator = vibrator
        self.vibrator.vibrate( duration )




# coding:utf8
__doc__ = 'Fluke 5520A multi-product calibrator'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Instrument
from .. import Env
import time
import re


class CommandSequenceError( Exception ):
    pass

class CommandOperationError( Exception ):
    pass

# eg: Fault 1300: Bad syntax (REM)
PATTEN_FAULT = re.compile('Fault ([0-9]+): .+')


class F5520A( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'F5520A'
    DEFAULT_TIMEOUT = 10
    DEFAULT_PROMPTS = re.compile( '([0-9]+?)> ' )
    DEFAULT_IDN = re.compile( 'FLUKE,(5520A|5522A),.+?' )
    promptIndex = None

    def checkReturnedPrompt( self, ret ):
        # check prompt if sequence is continuous
        match = self.DEFAULT_PROMPTS.match( ret[-1] )
        indexRet = int(match.group(1))
        if self.promptIndex is None:
            self.promptIndex = 2  # Ctrl-C will clear the sequence number
        if indexRet != self.promptIndex:
            err = 'sequence should be %d but returned %d'% (self.promptIndex, indexRet)
            raise CommandSequenceError(err)
        self.promptIndex += 1

    def connect( self, check_idn=True ):
        Instrument.SerialInstrument.connect( self, check_idn )
        self.setStandby()
        self.setRemote( True )

    def checkFaultLine( self, line ):
        match = re.match( PATTEN_FAULT, line )
        if match:
            #raise CommandOperationError( match.group(1))  # error code only
            raise CommandOperationError( line.strip() )  # complete message

    def isOperationCompleted( self ):
        ret = self.writeCommand( '*opc?' )
        if len(ret) >= 1:
            self.checkFaultLine( ret[0] ) 
        return ret[-1].strip() == '1'

    def waitForOperationCompleted( self, timeout=5 ):
        # wait for last command is completed
        # NOTE: command *wai does this without timeout mechanism 
        t0 = time.time()
        while True:
            if self.isOperationCompleted():
                break
            if time.time() > t0 + timeout:
                raise CommandOperationError()

    def setStandby( self ):
        # set standby mode
        self.writeCommand( 'STBY' )
        self.waitForOperationCompleted()

    def setOutput( self, mode ):
        # set output mode
        self.writeCommand( '*CLS' )
        ret = self.writeCommand( 'OUT ' + mode )
        if len(ret) >= 1:
            print( ret )
            self.checkFaultLine( ret[-1] )
        self.writeCommand( 'OPER' )
        self.waitForOperationCompleted()

    def getOutput( self ):
        # get output mode
        ret = self.writeCommand( 'OUT?' )
        return ret[0]

    def setRemote( self, remote=True ):
        # set remote/local mode
        self.writeCommand( 'remote' if remote else 'local' )
    
    def setLocal( self ):
        # set local mode
        self.setRemote( False )

    def stop( self ):
        # stop safely
        self.setStandby()
        self.setOutput( '0 V, 0 Hz' )

    def rtcSync( self ):
        # update to current time
        t = time.localtime()
        cmd = 'CLOCK %d,%d,%d,%d,%d,%d'% (t.tm_year, t.tm_mon, t.tm_mday,
                                          t.tm_hour, t.tm_min, t.tm_sec )
        self.writeCommand( cmd )

    def getRange( self ):
        # query range
        self.writeCommand( 'RANGE?' )
        return ret[0]

    def outputVolt( self, volt=0.0, amp=None, freq=None ):
        cmd = '%s V'% volt
        if amp is not None:
            cmd += ', %s A'% amp
        if freq is not None:
            cmd += ', %s Hz'% freq
        self.setOutput( cmd )

    def outputOhm( self, ohm=100.0 ):
        cmd = '%s Ohm'% ohm
        self.setOutput( cmd )





# coding:utf8
__doc__ = 'USART-HMI from Taojingchi Co.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
import re
import binascii
import time
from .. import Env, Instrument


RETURN_CODES = {
    0x00 : 'invalid',
    0x01 : 'ok',
    0x02 : 'control ID err',
    0x03 : 'page ID err',
    0x04 : 'image ID err',
    0x05 : 'font ID err',
    0x11 : 'baudrate err',
    0x12 : 'curve control ID err or channel err',
    0x1A : 'var name err',
    0x1B : 'oper err',
    0x1C : 'set err',
    0x1D : 'EEPROM err',
    0x1E : 'parm num err',
    0x1F : 'IO err',
    0x20 : 'escape char err',
    0x23 : 'var name too long',
    0x24 : 'usart buf overflow',
    0x65 : '',
    0x66 : '',
    0x67 : '',
    0x68 : '',
    0x70 : '',
    0x86 : '',
    0x87 : '',
    0x88 : '',
    0x89 : '',
    0xFD : '',
    0xFE : '',
    }


class TjcHMI( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'TjcHMI'
    DEFAULT_TERMINATOR_WRITE = '\xFF\xFF\xFF'
    DEFAULT_TERMINATOR_RESET = '\xFF\xFF\xFFbkcmd=3\xFF\xFF\xFF'
    DEFAULT_PROMPTS = re.compile( '[^\xFF]+\xff\xff\xff' )
    DEFAULT_CHECK_RETURN_COMMAND = False
    
    def readUntilPrompts( self, check_return_code=True ):
        '''read until prompts'''
        contents = []
        while True:
            byte = self.port.read(1)
            if byte:
                if Env.PYTHON_V3:
                    byte = chr(ord(byte))
                if (len(contents)==0) and (byte=='\xff'):
                    continue  # ignore leading 0xff
                else:
                    contents.append( byte )
            else:
                raise Instrument.CommandTimeoutError( binascii.hexlify(''.join(contents)) )
            ret = ''.join(contents)
            match = self.prompts.match( ret )
            if match:
                self.logger.debug( ret )
                if check_return_code:
                    assert ord(contents[0]) in RETURN_CODES.keys()
                return [''] + list(ret.rstrip('\xff')) + ['']

    def page( self, pageid=0 ):
        cmd = 'page %s'% str(pageid)
        self.writeCommand(cmd)

    def ref( self, obj ):
        cmd = 'ref %s'% str(obj)
        self.writeCommand(cmd)

    def click( self, click ):
        cmd = 'click %s'% str(click)
        self.writeCommand(cmd)

    def get( self, att ):
        cmd = 'get %s'% str(att)
        return self.writeCommand(cmd)

    #def print( self, att ):
    #    cmd = 'print %s'% str(att)
    #    return self.writeCommand(cmd)

    def vis( self, obj, state ):
        cmd = 'vis %s,%s'% [str(obj), str(int(state))]
        self.writeCommand(cmd)

    def sendme( self ):
        cmd = 'sendme'
        return self.writeCommand(cmd)

    def cls( self, color='black' ):
        cmd = 'cls %s'% str(color).upper()
        self.writeCommand(cmd)

    def line( self, x0, y0, x1, y1, color='white' ):
        cmd = 'line %s,%s,%s,%s,%s'% (x0, y0, x1, y1, str(color).upper())
        self.writeCommand(cmd)
    
    def cir( self, x, y, r, color='white' ):
        cmd = 'cir %s,%s,%s,%s'% (x, y, r, str(color).upper())
        self.writeCommand(cmd)

    def set( self, key, val ):
        if isinstance(val, str) or isinstance(val, unicode):
            cmd = '%s=\"%s\"'% (str(key), val.encode('gbk'))
        else:
            cmd = '%s=%s'% (str(key), str(val))
        self.writeCommand(cmd)

    def xstr( self, x,y,w,h,fontid,pointcolor,backcolor,xcenter,ycenter,sta,string ):
        cmd = 'xstr %s,%s,%s,%s,%s,%s,%s,%s,%s,%s,\"'% (str(x),str(y),str(w),str(h),
                    str(fontid),str(pointcolor),str(backcolor),str(xcenter),str(ycenter), str(sta))
        if Env.PYTHON_V3:
            cmd += ''.join( map(chr, string.encode(encoding='gbk')) )
        else:
            cmd += unicode(string).encode(encoding='gbk')
        cmd += '\"'    
        self.writeCommand(cmd)





# coding: utf8
__doc__ = 'SSD1306 controlled OLED12864 display'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush
from . import Bitmap
from . import Font
from . import Canvas as _Canvas
import time


class Ssd1306_Common():
    current_page = None
    current_x = None

    def cmd( self, b ):
        raise NotImplementedError

    def dat( self, b ):
        raise NotImplementedError
    
    def fill( self, pattern ):
        # fill full screen with fixed pattern (initialize)
        for p in range((self.height-1)/8+1):
            self.cmd(0xb0+p)
            self.cmd(0x01)
            self.cmd(0x10)
            for x in range(self.width):
                self.dat(pattern)
    
    def write_mem( self, page, x, mem ):
        # write display buffer at specified page/x
        # optimize: do not re-send position if it's continuous
        if page != self.current_page:
            self.cmd(0xb0+page)
            self.current_page = page 

        if x != self.current_x:
            self.cmd(((x&0xf0)>>4)|0x10)
            self.cmd((x&0x0f)|0x00)
            self.current_x = x
           
        self.dat( mem )
        self.current_x += 1

    def init(self):
        # registers initialize, refer to datasheet
        self.cmd(0xae)   # turn off oled panel
        self.cmd(0x00)   # set low column address
        self.cmd(0x10)   # set high column address
        self.cmd(0x40)   # set start line address  Set Mapping RAM Display Start Line (0x00~0x3F)
        self.cmd(0x81)   # set contrast control register
        self.cmd(0xcf)   # Set SEG Output Current Brightness
        self.cmd(0xa1)   # Set SEG/Column Mapping
        self.cmd(0xc8)   # Set COM/Row Scan Direction
        self.cmd(0xa6)   # set normal display
        self.cmd(0xa8)   # set multiplex ratio(1 to 64)
        self.cmd(0x3f)   # 1/64 duty
        self.cmd(0xd3)   # set display offset Shift Mapping RAM Counter (0x00~0x3F)
        self.cmd(0x00)   # not offset
        self.cmd(0xd5)   # set display clock divide ratio/oscillator frequency
        self.cmd(0x80)   # set divide ratio, Set Clock as 100 Frames/Sec
        self.cmd(0xd9)   # set pre-charge period
        self.cmd(0xf1)   # Set Pre-Charge as 15 Clocks & Discharge as 1 Clock
        self.cmd(0xda)   # set com pins hardware configuration
        self.cmd(0x12)   # 
        self.cmd(0xdb)   # set vcomh
        self.cmd(0x40)   # Set VCOM Deselect Level
        self.cmd(0x20)   # Set Page Addressing Mode (0x00/0x01/0x02)
        self.cmd(0x02)   # 
        self.cmd(0x8d)   # set Charge Pump enable/disable
        self.cmd(0x14)   # set(0x10) disable
        self.cmd(0xa4)   # Disable Entire Display On (0xa4/0xa5)
        self.cmd(0xa6)   # Disable Inverse Display On (0xa6/a7) 
        self.cmd(0xaf)   # turn on oled panel


        
class Ssd1306_I2C(Ssd1306_Common):

    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x3C, scl, sda, delay=5 )
        self.init()

    def cmd( self, b ):
        self.controller.i2c( [0x00, b] )
 
    def dat( self, b ):
        self.controller.i2c( [0x40, b] )
    
    def fill( self, pattern ):
        # optimized code for i2c command: multi data write
        for p in range((self.height-1)/8+1):
            self.cmd(0xb0+p)
            self.cmd(0x01)
            self.cmd(0x10)
            for x in range(self.width/8):
                self.controller.i2c( [
                    0xC0, pattern, 0xC0, pattern, 0xC0, pattern, 0xC0, pattern, 
                    0xC0, pattern, 0xC0, pattern, 0xC0, pattern, 0xC0, pattern, 
                    ] )
 

class Ssd1306_SPI(Ssd1306_Common):

    def __init__( self, controller, rst=None, dc=None, sdo=None, sdi=None, sck=None, cs=None ):
        self.controller = controller
        self.controller.spi_init( sdo=sdo, sdi=sdi, sck=sck, cs=cs, delay=5 )
        self.rst_pin = rst
        self.dc_pin = dc
        self.dc = None
        self.controller.pinOutputHigh( [rst, dc] )
        self.reset()
        self.init()

    def reset( self ):
        self.controller.pinClr( self.rst_pin )
        time.sleep(0.1)
        self.controller.pinSet( self.rst_pin )
     
    def cmd( self, d ):
        if self.dc != 0:
            self.controller.pinClr( self.dc_pin )
            self.dc = 0
        self.controller.spi( [d] )
 
    def dat( self, d ):
        if self.dc != 1:
            self.controller.pinSet( self.dc_pin )
            self.dc = 1
        self.controller.spi( [d] )
    
   

class OLED128x64_I2C(Ssd1306_I2C):
    width, height = 128, 64


class OLED128x64_SPI(Ssd1306_SPI):
    width, height = 128, 64



# abstract canvas based on matrix led controller
class Canvas(_Canvas.Canvas):

    def __init__( self, display ):
        self.display = display
        self.width = display.width
        self.height = display.height
        self.pages = (display.height-1)/8+1
        self.buffer = [ [0] * self.width for i in range(self.pages) ]
        self.display.fill( 0 )
        self.dirty = [0] * self.pages

    def flush( self, force=False ):
        for p in range(self.pages):
            for x in range(self.width):
                if force or (self.dirty[p] & (1<<x)):
                    self.display.write_mem( p, x, self.buffer[p][x] )
                    self.dirty[p] &= ~(1<<x);

    def setPixel( self, x, y, color=None, flush=True ):
        # check if it's out of range
        if x < 0 or x >= self.width:
            return
        if y < 0 or y >= self.height:
            return
        page = y/8
        byte_mask = 1 << (y%8)
        byte_index = x
        if color is None:
            color = 1
        if color:
            self.buffer[page][byte_index] |= byte_mask
        else:
            self.buffer[page][byte_index] &= ~byte_mask
        self.dirty[page] |= 1<<x
        if flush:
            self.flush()
    




# coding: utf8
__doc__ = 'Dallas 1-wire bus controller'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import crcmod
import crcmod.predefined


class Dallas1W():
        
    def __init__( self, controller, pin=None ):
        self.controller = controller
        self.crc = crcmod.predefined.mkPredefinedCrcFun('crc-8-maxim')
        if pin:
            self.controller.writeCommand('ds1w -I -p%s'% pin)
        else:
            self.controller.writeCommand('ds1w -I')

    def readBit( self ):
        r = self.controller.writeCommand('ds1w -z')
        return int(r[0])

    def read( self, num=1 ):
        if num == 1:
            return int(self.controller.writeCommand('ds1w -r')[0])
        else:
            r = []
            for i in range(num):
                r.append( int(self.controller.writeCommand('ds1w -r')[0]) )
            return r

    def write0( self ):
        self.controller.writeCommand('ds1w -0') 

    def write1( self ):
        self.controller.writeCommand('ds1w -1') 

    def writeBit( self, bit ):
        if bit:
            self.write1()
        else:
            self.write0()

    def write( self, byte ):
        self.controller.writeCommand('ds1w -w%d'% byte) 

    def reset( self ):
        self.controller.writeCommand('ds1w -R') 

    def scan(self):
        result = []
        path = []
        branch = []
        while True:
            self.reset()
            self.write( 0xF0 )
            for b in path:
                self.readBit()
                self.readBit()
                self.writeBit(b)
            depth = len(path)
            for i in range(len(path), 64):
                a = self.readBit()
                b = self.readBit()
                if (a==0) and (b==0):  # both 0/1
                    branch.append(i)
                    path.append(0)
                    self.writeBit(0)  # search the left branch
                    depth += 1
                elif (a==0) or (b==0):  # only 0/1
                    path.append(a)
                    self.writeBit(a)
                    depth += 1
                else:  # no device
                    break
            if depth == 64:  # found
                r = []
                for i in range(8):
                    v = 0
                    for j in range(8):
                        if path[i*8+j]:
                            v |= 1<<j
                    r.append(v)
                result.append(r)
            if branch:
                b = branch.pop()
                path = path[:b]
                path.append(1)  # search the right branch
            else:  # stop
                break
        # sn filter 
        sn = []
        for r in result:
            if r[0] != self.FAMILY_CODE:
                continue
            crc = self.crc(''.join([chr(i) for i in r[:-1]]))
            if crc == r[-1]:
                sn.append( r[1:-1] )
                sn[-1].reverse()
        return [''.join(['%02X'%i for i in j]) for j in sn]
                 
    def readRom( self ):
        # ignore multi-slave, read only the first one
        self.reset()
        self.write( 0x33 )
        r = self.read(8)
        crc_calc = self.crc(''.join([chr(i) for i in r[:7]]))
        if r[7] != crc_calc:
            code = '-'.join(['%02X'%i for i in r])
            raise Exception("Crc error:%s"% code )
        if r[0] != self.FAMILY_CODE:
            raise Exception("Family code (0x%X) not match: 0x%X"% (self.FAMILY_CODE, r[0]))
        serial_num = r[1:7]
        serial_num.reverse()
        return ''.join(['%02X'%i for i in serial_num])

    getSerialNumber = readRom

    def scanSerialNumber( self ):
        return self.scan()

    def writeROM( self, sn ):
        self.write( self.FAMILY_CODE )
        sn2i = [int('0x'+sn[i*2:i*2+2], 16) for i in range(5,-1,-1)]
        for i in sn2i:
            self.write( i )
        crc = self.crc(''.join([chr(i) for i in [self.FAMILY_CODE]+sn2i]))
        self.write( crc )



class DS18B20( Dallas1W ):
    FAMILY_CODE = 0x28
    parasite = None

    def getTemperature( self, sn=None ):
        # read temperature from scratchpad, ignore remaining (including CRC)
        self.reset()
        if sn is not None:
            self.write( 0x55 )
            self.writeROM( sn )
        else:
            self.write( 0xCC )
        self.write( 0xBE )
        t = Utils.s2h(chr(self.read()) + chr(self.read()))
        return round(t*0.0625, 1)

    def getScratchpad( self, sn=None ):
        # read all from scratchpad
        self.reset()
        if sn is not None:
            self.write( 0x55 )
            self.writeROM( sn )
        else:
            self.write( 0xCC )
        self.write( 0xBE )
        r = self.read(9)
        t = Utils.s2h(chr(r[0]) + chr(r[1]))
        th = Utils.s2b(chr(r[2]))
        tl = Utils.s2b(chr(r[3]))
        res = 9 + ((Utils.s2B(chr(r[4])) >> 5) & 0x03)
        crc_calc = self.crc(''.join([chr(i) for i in r[:8]]))
        if r[8] != crc_calc:
            raise Exception("Crc error")
        return [round(t*0.0625, 1), th, tl, res]

    getAll = getScratchpad

    def setScratchpad( self, sn=None, th=0, tl=0, res=12 ):
        if res not in [9, 10, 11, 12]:
            raise Exception("Resolution must between 9~12")
        self.reset()
        if sn is not None:
            self.write( 0x55 )
            self.writeROM( sn )
        else:
            self.write( 0xCC )
        self.write( 0x4E )
        self.write( th & 0xFF )
        self.write( tl & 0xFF )
        self.write( ((res-9) & 0x3) << 5 )
    
    config = setScratchpad
 
    def convert( self, sn=None ):
        self.reset()
        if sn is not None:
            self.write( 0x55 )
            self.writeROM( sn )
        else:
            self.write( 0xCC )
        self.write( 0x44 )

    def checkParasite( self, sn=None ):
        self.reset()
        if sn is not None:
            self.write( 0x55 )
            self.writeROM( sn )
        else:
            self.write( 0xCC )
        self.write( 0xB4 )
        self.parasite = bool(self.readBit()==0)
        return self.parasite
            

class DS1990( Dallas1W ):
    FAMILY_CODE = 0x01
 



# coding: utf8
__doc__ = 'MAX30102 pulse oximeter and hart-rate sensor'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'


REG_INTR_STATUS_1 = 0x00
REG_INTR_STATUS_2 = 0x01
REG_INTR_ENABLE_1 = 0x02
REG_INTR_ENABLE_2 = 0x03
REG_FIFO_WR_PTR = 0x04
REG_OVF_COUNTER = 0x05
REG_FIFO_RD_PTR = 0x06
REG_FIFO_DATA = 0x07
REG_FIFO_CONFIG = 0x08
REG_MODE_CONFIG = 0x09
REG_SPO2_CONFIG = 0x0A
REG_LED1_PA = 0x0C
REG_LED2_PA = 0x0D
REG_MULTI_LED_CTRL1 = 0x11
REG_MULTI_LED_CTRL2 = 0x12
REG_TEMP_INTR = 0x1F
REG_TEMP_FRAC = 0x20
REG_TEMP_CONFIG = 0x21
REG_REV_ID = 0xFE
REG_PART_ID = 0xFF



class Max30102():
        
    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x57, scl, sda )
        self.reset()
        self.init()

    def readReg( self, addr ):
        return self.controller.i2c( [addr], read=1 )[0]

    def writeReg( self, addr, dat ):
        self.controller.i2c( [addr, dat] )

    def reset( self ):
        self.writeReg( REG_MODE_CONFIG, 0x40 )

    def init( self ):
        self.writeReg( REG_INTR_ENABLE_1, 0xc0 ) # INTR setting
        self.writeReg( REG_INTR_ENABLE_2, 0x00 )
        self.writeReg( REG_FIFO_WR_PTR, 0x00 )  # FIFO_WR_PTR[4:0]
        self.writeReg( REG_OVF_COUNTER, 0x00 )  # OVF_COUNTER[4:0]
        self.writeReg( REG_FIFO_RD_PTR, 0x00 )  # FIFO_RD_PTR[4:0]
        self.writeReg( REG_FIFO_CONFIG, 0X0F )  # sample avg: 1, fifo rollover: 0, fifo almost full: 17
        self.writeReg( REG_MODE_CONFIG, 0x03 )  # 0x02 for Red only, 0x03 for SpO2 mode 0x07 multimode LED
        self.writeReg( REG_SPO2_CONFIG, 0x27 )  # SPO2_ADC range: 4096nA,  SPO2 sample rate (100 Hz), LED pulseWidth (400uS)
        self.writeReg( REG_LED1_PA, 0x24 )  # 7mA for LED1(RED)
        self.writeReg( REG_LED2_PA, 0x24 )  # 7mA for LED2(IRED)

    def readStatus( self ):
        s1 = self.readReg( REG_INTR_STATUS_1 )
        s2 = self.readReg( REG_INTR_STATUS_2 )
        return s1, s2

    def readFifo( self ):
        self.readStatus()
        r = self.controller.i2c( [REG_FIFO_DATA], read=6 )
        red = (r[0]<<16) + (r[1]<<8) + r[2]
        ir = (r[3]<<16) + (r[4]<<8) + r[5]
        return red, ir




# coding: utf8
__doc__ = 'utilities'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from os import system, remove
from sys import platform, stdout
from binascii import hexlify, unhexlify, crc32
from random import randint
from subprocess import Popen, PIPE
from struct import pack, unpack
from time import strftime, localtime
import json
from . import Env
import traceback


MAX_LINE_LEN = 80

def writeTitle( title='', mark='=', color=True ):
    if Env.PYTHON_V3:
        title = str(title)
    else:
        title = unicode(title)
    line = title.join( '  ' ) if title.strip() else ''
    marks = mark + mark
    while len(line) < MAX_LINE_LEN:
        line = line.join( marks )
    line = line[:MAX_LINE_LEN]
    if color:
        line = colored_string(line, 'black', 'white', style='bold')
    if Env.PYTHON_V3:
        stdout.write( str(line) + '\n' )
    else:
        stdout.write( unicode(line) + '\n' )
    stdout.flush()

def generateRandomFile( size ):
    FILE = open( Env.TEST_WR, 'w+' )
    counter = 0
    while counter < size:
        FILE.write( chr(randint(0, 255)) )
        counter += 1
    FILE.close()    

def diffFiles( file1, file2 ):
    if Env.platform == 'linux2':
        ret = system( 'colordiffbin %s %s'% (file1, file2) )
        return not bool(ret)
    else:
        import filecmp
        return filecmp.cmp( file1, file2 )

def diffTestFiles():
    return diffFiles( Env.TEST_WR, Env.TEST_RD )

def testSingleRandomFile( device, targetfile, size ):
    generateRandomFile( size )
    device.putFile( targetfile, Env.TEST_WR )
    device.getFile( targetfile, Env.TEST_RD )
    return diffTestFiles()

def dumpFile( fname ):
    system( "hexdump -C %s"% fname )

def dumpMem( mem, method=2 ):
    if platform == 'win32':
        method = 1  # force
    if method == 1:
        # method 1: use binascii
        stdout.write( hexlify(mem) + '\n' )
        stdout.flush()
    elif method == 2:
        # method 2: use hexdump
        proc = Popen( ['hexdump -C'], stdin = PIPE, stdout = PIPE, shell=True )
        proc.stdin.write( mem )
        proc.stdin.flush()
        proc.stdin.close()
        read = proc.stdout.read() 
        stdout.write( read + '\n' )
        stdout.flush()


def parseKeyValueLines( lines ):
    # parse lines in format:
    #   key1: val1
    #   key2:       <-- treat as empty string 
    #   key3: val3
    #   ...
    # return k-v dictionary, evaluate k/v as trimmed case-sensitive string
    # if two or more keys are the same, preserve the last one
    var = {}
    for l in lines:
        idx = l.find(': ')
        if idx <= 0:
            continue
        a = l[:idx]
        b = l[idx+2:]
        var[a.strip()] = b.strip()
    return var


def enumPorts():
    ports = []
    if platform == 'linux2':
        from glob import glob
        ports += glob( '/dev/ttyACM*' )
        ports += glob( '/dev/ttyUSB*' )
    elif platform == 'win32':
        try:
            from serial.tools import list_ports
            for p in list_ports.comports():
                ports.append( p[0] )
        except ImportError:
            pass 
    return ports


def enumPortsByClass( cls, callback=None ):
    from . import Instrument
    ret = []
    for port in enumPorts():
        try:
            obj = cls( port )
            if callback:
                callback( obj )
            ret.append(port)
        except Instrument.PortNotFound:
            pass
        except Instrument.IDNMatchError:
            pass
        except Instrument.CommandTimeoutError:
            pass
        except Instrument.CommandSyntaxError:
            pass
        except Instrument.CommandExecuteError:
            pass
        except Instrument.ResponseError:
            pass
        except Instrument.UnknownPortError:
            pass
        finally:
            try:
                obj.disconnect()
                del obj
            except NameError:
                pass
    return ret


def espeak( contents ):
    if Env.platform == 'linux2':
        system( 'espeak %s'% contents )


# sign convert
def i2I( val ):
    return unpack('I', pack('i', val)) [0]

def I2i( val ):
    return unpack('i', pack('I', val)) [0]

def h2H( val ):
    return unpack('H', pack('h', val)) [0]

def H2h( val ):
    return unpack('h', pack('H', val)) [0]

# float convert
def i2f( val ):
    return unpack('f', pack('i', val)) [0]

def I2f( val ):
    return unpack('f', pack('I', val)) [0]

def f2i( val ):
    return unpack('i', pack('f', val)) [0]
 
def f2I( val ):
    return unpack('I', pack('f', val)) [0]

# memory convert 
def s2f( val ):
    return unpack('f', val) [0]

def s2d( val ):
    return unpack('d', val) [0]

def s2b( val ):
    return unpack('B', val) [0]
 
def s2B( val ):
    return unpack('B', val) [0]
 
def s2h( val ):
    return unpack('h', val) [0]
 
def s2H( val ):
    return unpack('H', val) [0]
 
def s2i( val ):
    return unpack('i', val) [0]
 
def s2I( val ):
    return unpack('I', val) [0]
 
def s2q( val ):
    return unpack('q', val) [0]
 
def s2Q( val ):
    return unpack('Q', val) [0]
    
def f2s( val ):
    return pack('f', val)

def d2s( val ):
    return pack('d', val)

def i2s( val ):
    return pack('i', val)

def I2s( val ):
    return pack('I', val)

def h2s( val ):
    return pack('h', val)

def H2s( val ):
    return pack('H', val)

def b2s( val ):
    return pack('b', val)

def B2s( val ):
    return pack('B', val)

def q2s( val ):
    return pack('q', val)

def Q2s( val ):
    return pack('Q', val)

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
def DATETIME():
    return strftime(DATETIME_FORMAT)

def I2t( val, format=None ):
    if format is None:
        format = DATETIME_FORMAT
    return strftime(format, localtime(val))

def dt2s( dt, need_ms=False ):
    hour = int(dt / 3600)
    dt -= hour*3600
    minute = int(dt / 60)
    dt -= minute * 60
    sec = int(dt)
    if need_ms:
        ms = (dt - sec) * 1000
        return '%d:%02d:%02d.%03d'% (hour, minute, sec, ms)
    else:
        return '%d:%02d:%02d'% (hour, minute, sec)



def pyobfuscate( pyin, pyout, merged_modules=[], tmpfile=None, remove_tmp=True ):
    if Env.platform != 'linux2':
        print( 'must run under linux' )
        return 
    # prepare source file with specified modules merged
    src = open(pyin,'r').read()
    for m in merged_modules:
        pattern = 'from %s import *'% m
        replace = open('%s.py'% m,'r').read()
        src = src.replace( pattern, replace )
    if tmpfile is None:
        tmpfile = Env.TEST_WR
    open(tmpfile,'w+').write(src)
    # obfuscate
    system( 'pyobfuscate %s > %s'% (tmpfile, pyout) )
    if remove_tmp:
        remove(tmpfile) 

def compact_cjson_str(obj):
    LINE_LIMIT = 128
    cfg = list(json.dumps(obj,separators=(',',':')))
    lst = []
    while cfg:
        if len(cfg) <= LINE_LIMIT:
            lst.append(''.join(cfg))
            break
        p = LINE_LIMIT - 1
        while cfg[p] != ',':
            p -= 1
        lst.append(''.join(cfg[:p+1]))
        cfg = cfg[p+1:]
    #print lst 
    return lst


def STR(s):
    try:
        s = _(s)
    except:
        return s
    try:
        return s.decode('utf8')
    except:
        return s


def crc( data, oldcrc=0 ):
    v = crc32( data, oldcrc )
    return v & 0xFFFFFFFF


class colored_string():
    COLOR = {
        'black': 30, 'red': 31, 'green': 32, 'yellow': 33, 'blue': 34,
        'magenta': 35, 'cyan': 36, 'white': 37, 'reset': 39, 
        'lightblack': 90, 'lightred': 91, 'lightgreen': 92,
        'lightyellow': 93, 'lightblue': 94, 'lightmagenta': 95,
        'lightcyan': 96, 'lightwhite': 97 }
    STYLE = {
        'bold': 1, 'faint': 2, 'italic': 3, 'underline': 4, 'blink': 5, 
        'reverse': 7, 'concealed': 8, 'crossed': 9, 'normal': 22, }
    def __init__( self, string, fgcolor=None, bgcolor=None, style=None ):
        self.string = string
        self.fgcolor = fgcolor
        self.bgcolor = bgcolor
        self.style = style

    def __str__( self ):
        head = ''
        if self.style is not None:
            head += '\x1b[%dm'% (self.STYLE[self.style])
        if self.bgcolor is not None:
            head += '\x1b[%dm'% (self.COLOR[self.bgcolor]+10)
        if self.fgcolor is not None:
            head += '\x1b[%dm'% self.COLOR[self.fgcolor]
        if head:
            return self.string.join([head, '\x1b[0m'])
        else:
            return self.string

def red_string(string):
    return colored_string(string, 'red')

def green_string(string):
    return colored_string(string, 'green')

def blue_string(string):
    return colored_string(string, 'blue')

def yellow_string(string):
    return colored_string(string, 'yellow')

def magenta_string(string):
    return colored_string(string, 'magenta')

def cyan_string(string):
    return colored_string(string, 'cyan')

def bold_string(string):
    return colored_string(string, style='bold')

def faint_string(string):
    return colored_string(string, style='faint')

RED = red_string
GREEN = green_string
BLUE = blue_string
YELLOW = yellow_string
MAGENTA = magenta_string
CYAN = cyan_string
BOLD = bold_string
FAINT = faint_string

def colored_log( log_string ):
    # parse log type and colorized it
    items = log_string.split(' ')
    items_len = len(items)
    if (items_len >= 4) and (len(items[2]) == 1):
        # format: 20YY-M-D HH:MM:SS T module: msg
        TYPE = items[2] 
    elif (items_len >= 3) and (len(items[1]) == 1):
        # format: HH:MM:SS.ms T module: msg
        TYPE = items[1] 
    else:
        return colored_string(log_string, 'white', 'red', 'bold')
    if TYPE in ['E']:
        return colored_string(log_string, 'lightred')
    elif TYPE in ['W']:
        return colored_string(log_string, 'lightyellow')
    elif TYPE in ['I']:
        return log_string
    elif TYPE in ['D']:
        return faint_string(log_string)
    else:
        return colored_string(log_string, 'white', 'red', 'bold')
        

def check_traceback_for_errline(e):
    if Env.DEBUG:
        traceback.print_exc(stdout)
    err = []
    for line in traceback.format_exc().splitlines():
        line = line.rstrip()
        if line:
            err.append(line)
    err_info = err.pop()
    err_file, err_line, err_module = None, None, None
    while err:
        l = err.pop()
        try:
            if not l.startswith('  File '):
                continue
        except UnicodeDecodeError:
            continue
        try:
            _file, _line, _module = l.split(', ')
        except ValueError:
            _file, _line = l.split(', ')
            _module = '<module>'
        err_module = _module.lstrip('in ')
        if err_module == '<module>':
            err_file = _file.lstrip('  File ')
            err_line = int(_line.lstrip('line '))
            break
    return err_line



class FCFS:
    class Signature():
        def __init__( self, info ):
            self.info = info
        def __len__( self ):
            return len(self.info)
        def __str__( self ):
            return self.info
    
    class Node():
        def __init__( self, file ):
            self.file = file
            self.offset = 0
            self.size = file.size
        def __len__( self ):
            return 4
        def __str__( self ):
            return pack('H', self.offset) + pack('H', self.size)
    
    class File():
        def __init__( self, fname, contents ):
            self.fname = fname
            self.contents = contents
            self.size = len(self.contents)
        def __len__( self ):
            return len(self.fname.encode('utf8')) + 1 + self.size
        def __str__( self ):
            return self.fname.encode('utf8') + '\00' + self.contents

    def __init__( self ):
        self.reset()

    def reset( self ):
        self.head = [self.Signature('FCFS')]
        uid = [chr(randint(0,255)) for i in range(12)]
        self.uid = self.Signature(''.join(uid))
        self.head.append( self.uid )
        self.contents = []

    def setUID( self, uid ):
        while len(uid) < 12:
            uid += '\x00'
        self.uid.info = uid[:12]

    def setIntegerUID( self, integer, msb=True ):
        uid = I2s(integer&0xFFFFFFFF)
        uid += I2s((integer>>32)&0xFFFFFFFF)
        uid += I2s((integer>>64)&0xFFFFFFFF)
        if msb:
            uid = list(uid)
            uid.reverse()
            self.setUID( ''.join(uid) )
        else:
            self.setUID( uid )
 
    def appendSignature( self, item ):
        self.contents.append( item )

    def appendFile( self, name, contents ):
        f = self.File(name, contents)
        self.contents.append( f )
        self.head.append( self.Node(f) )

    def generate( self ):
        self.head.append( self.Signature('\x00\x00\x00\x00') )  # end node
        self.head.append( self.Signature(strftime('<%y%m%d%H%M%S>')) )  # timestamp
        # re-calculate the offsets
        offset_counter = sum([len(h) for h in self.head])
        for idx in range(len(self.contents)):
            f = self.head[idx+2]
            f.offset = offset_counter
            offset_counter += len(f.file)
        self.contents.append( self.Signature('END') )  # end
        output = self.head + self.contents
        output = ''.join([str(o) for o in output])
        return output
 








# coding: utf8
__doc__ = 'max7219 matrix led controller module'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush
from . import Bitmap
from . import Font
from . import Canvas as _Canvas
from . import SegmentTable

ADDR_NOP = 0
ADDR_D0 = 1
ADDR_D1 = 2
ADDR_D2 = 3
ADDR_D3 = 4
ADDR_D4 = 5
ADDR_D5 = 6
ADDR_D6 = 7
ADDR_D7 = 8
ADDR_DECODE_MODE = 9
ADDR_INTENSITY = 10
ADDR_SCAN_LIMIT = 11
ADDR_SHUTDOWN = 12
ADDR_DISP_TEST = 15



class LED():

    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None, intensity=None ):
        self.controller = controller
        self.controller.spi_init( sdi, sdo, sck, cs, width=16, delay=10 )
        self.controller_num = (int((self.width-1)/8)+1) * (int((self.height-1)/8)+1)
        if intensity is None:
            intensity = 0x08
        self.intensity = intensity & 0x0F  # range: 0 ~ 0x0F
        self.reset()

    def write(self, addr, dat):
        if isinstance(dat, list):
            # multiple controllers, different value
            val = [((addr & 0x0F)<<8) + (d & 0xFF) for d in dat]
            val.reverse()
            self.controller.spi( val )
        elif self.controller_num > 1 :
            # multiple controllers, same value
            val = ((addr & 0x0F)<<8) + (dat & 0xFF)
            self.controller.spi( [val] * self.controller_num )
        else:
            # single controller
            val = ((addr & 0x0F)<<8) + (dat & 0xFF)
            self.controller.spi( [val] )

    def reset(self):
        self.write( ADDR_SHUTDOWN, 1 )
        self.write( ADDR_SCAN_LIMIT, 7 )
        self.write( ADDR_DECODE_MODE, 0 )
        self.write( ADDR_INTENSITY, self.intensity )
        self.write( ADDR_D0, 0 )
        self.write( ADDR_D1, 0 )
        self.write( ADDR_D2, 0 )
        self.write( ADDR_D3, 0 )
        self.write( ADDR_D4, 0 )
        self.write( ADDR_D5, 0 )
        self.write( ADDR_D6, 0 )
        self.write( ADDR_D7, 0 )
        self.write( ADDR_DISP_TEST, 0 )

    def set_intensity( self, val ):
        self.write( ADDR_INTENSITY, val & 0x0F )
        self.intensity = val

    def test_all( self ):
        self.write( ADDR_D0, 0xFF )
        self.write( ADDR_D1, 0xFF )
        self.write( ADDR_D2, 0xFF )
        self.write( ADDR_D3, 0xFF )
        self.write( ADDR_D4, 0xFF )
        self.write( ADDR_D5, 0xFF )
        self.write( ADDR_D6, 0xFF )
        self.write( ADDR_D7, 0xFF )

    def write_line( self, index, line ):
        self.write( ADDR_D0 + 7 - index, line )

    

class LED8x8(LED):
    width, height = 8, 8

class LED16x8(LED):
    width, height = 16, 8

class LED24x8(LED):
    width, height = 24, 8

class LED32x8(LED):
    width, height = 32, 8
 
class LED64x8(LED):
    width, height = 64, 8
 
class LED8x16(LED):
    width, height = 8, 16

class LED16x16(LED):
    width, height = 16, 16

class LED24x16(LED):
    width, height = 24, 16

class LED32x16(LED):
    width, height = 32, 16

class LED8x24(LED):
    width, height = 8, 24

class LED16x24(LED):
    width, height = 16, 24

class LED24x24(LED):
    width, height = 24, 24

class LED32x24(LED):
    width, height = 32, 24

class LED8x32(LED):
    width, height = 8, 32

class LED16x32(LED):
    width, height = 16, 32

class LED24x32(LED):
    width, height = 24, 32

class LED32x32(LED):
    width, height = 32, 32



# abstract canvas based on matrix led controller
class Canvas(_Canvas.Canvas):

    def __init__( self, display ):
        if not isinstance(display, LED):
            raise Exception("display type error")
        self.display = display
        self.width = display.width
        self.height = display.height
        self.controller_num = display.controller_num
        self.buffer = [ [0] * self.controller_num for i in range(8) ]
        self.dirty = 0xFF 

    def flush( self, force=False ):
        for i in range(8):
            if force or (self.dirty & (1<<i)):
                self.display.write_line( i, self.buffer[i] )
                self.dirty &= ~(1<<i)

    def setPixel( self, x, y, color=None, flush=True ):
        # check if it's out of range
        if x < 0 or x >= self.width:
            return
        if y < 0 or y >= self.height:
            return
        # transform postion to chain mode
        while y >= 8:
            y -= 8 
            x += self.width
        byte_mask = 1 << (x%8)
        byte_index = int(x/8)
        if color is None:
            color = 1
        if color:
            self.buffer[y][byte_index] |= byte_mask
        else:
            self.buffer[y][byte_index] &= ~byte_mask
        self.dirty |= 1<<y
        if flush:
            self.flush()
    
    def clrScr( self, flush=True ):
        # optimized api
        for i in range(8):
            for j in range(self.controller_num):
                self.buffer[i][j] = 0 
        self.dirty = 0xFF
        if flush:
            self.flush()
 

#  
#      a
#    -----
#  f|     | b
#   |  g  |
#    ----- 
#  e|     | c
#   |     |
#    ----- .h
#      d
# D7 D6 D5 D4 D3 D2 D1 D0
# DP A  B  C  D  E  F  G 
SA = 0x40
SB = 0x20
SC = 0x10
SD = 0x08
SE = 0x04
SF = 0x02
SG = 0x01
SH = 0x80

SEGMENT_TABLE_HEX = SegmentTable.generateSegmentTableHEX(SA,SB,SC,SD,SE,SF,SG,SH)
SEGMENT_TABLE_ASCII = SegmentTable.generateSegmentTableASCII(SA,SB,SC,SD,SE,SF,SG,SH)

class LED8(LED):
    DIGITS = 8
       
    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None, intensity=None ):
        self.controller = controller
        self.controller.spi_init( sdi, sdo, sck, cs, width=16, delay=10 )
        self.controller_num = int((self.DIGITS-1)/8)+1
        if intensity is None:
            intensity = 0x08
        self.intensity = intensity & 0x0F  # range: 0 ~ 0x0F
        self.reset()

    def clear( self ):
        self.display_segs( 0, "    ", ascii_mode=True )

    def display_string( self, string, index=0 ):
        segments = []
        for c in string:
            if not c in ['.', ':']:
                segments.append( SEGMENT_TABLE_ASCII[ord(c)] )
            elif len(segments) > 0:
                segments[-1] |= SegmentTable.SH
        for i, s in enumerate(segments[:self.DIGITS], index):
            self.write_line( i, s )
 
    def display_integer( self, val, hex_mode=False ):
        if hex_mode:
            string = (('%%%dX'% self.DIGITS)% val)[-self.DIGITS:]
        else:
            string = (('%%%dd'% self.DIGITS)% val)[-self.DIGITS:]
        self.display_segs( 0, string[:self.DIGITS], ascii_mode=True )
    
    def display_segment( self, index, segment, ascii_mode=None, hex_mode=None ):
        if hex_mode:
            segment = SEGMENT_TABLE_HEX[segment]
        elif ascii_mode:
            segment = SEGMENT_TABLE_ASCII[ord(segment)]
        self.segments[index] = segment
        self.write_line( index, segment )

    def display_segments( self, index, segments, ascii_mode=None, hex_mode=None ):
        segments = segments[:self.DIGITS]
        if hex_mode:
            segments = [ SEGMENT_TABLE_HEX[c] for c in segments ]
        elif ascii_mode:
            segments = [ SEGMENT_TABLE_ASCII[ord(c)] for c in segments ]
        self.segments = segments
        for i,s in enumerate(segments,index):
            self.write_line( i, s )

    clrscr = clear
    display_str = display_string
    display_int = display_integer
    display_seg = display_segment
    display_segs = display_segments


 



# coding: utf8
__doc__ = 'Products from Shanghai Linkong Software Technologies Co., Ltd.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from . import ShellKit
from . import ShellLab
from . import VAP



# coding: utf8
__doc__ = 'ADS1113/4/5 16-bit ADC'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Instrument, Utils
import time


REG_CONV     = 0  # Conversion
REG_CONFIG   = 1  # Config
REG_THRES_L  = 2  # Lo_thres
REG_THRES_H  = 3  # Hi_thres

ADDR = [0x48, 0x49, 0x4A, 0x4B]
PGA_RANGE = [6.144, 4.096, 2.048, 1.024, 0.512, 0.256, 0.256, 0.256]

class Ads1115():

    def __init__( self, controller, scl=None, sda=None, addr_mode=0 ):
        self.controller = controller
        self.controller.i2c_init( ADDR[addr_mode], scl, sda )
        self.reset()
        self.config(mode=0)  # continuous measure mode

    def readReg( self, addr ):
        h, l = self.controller.i2c( [addr], read=2 )
        return (h<<8)+l

    def writeReg( self, addr, dat ):
        self.controller.i2c( [addr&0x03, dat>>8, dat&0xFF] )

    def reset( self ):
        # reset all chips on the bus that response to common address
        try:
            self.controller.i2c( [0x06], addr=0x0 )
        except Instrument.CommandExecuteError:
            pass  # device has no ACK response

    def config( self, os=0, mux=0, pga=2, mode=1, dr=4, comp_mode=0, comp_pol=0, comp_lat=0, comp_que=3 ):
        config = (os<<15)+(mux<<12)+(pga<<9)+(mode<<8)+(dr<<5)+ \
                 (comp_mode<<4)+(comp_pol<<3)+(comp_lat<<2)+(comp_que)
        self.writeReg( REG_CONFIG, config )
        self.pga_range = PGA_RANGE[pga]

    def setThreshold( self, lo=0x8000, hi=0x7FFF ):
        self.writeReg( REG_THRES_L, lo )
        self.writeReg( REG_THRES_H, hi )

    def readConv( self ):
        val = Utils.H2h( self.readReg( REG_CONV ) )
        return val * self.pga_range / (2**15)




# coding: utf8
__doc__ = 'TM7707 24-bit ADC'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import time


# Communication register
REG_COMM     = 0x00  # Communication register
REG_SETUP    = 0x10  # Setting register
REG_FILTER_H = 0x20  # Filter(high) register
REG_DATA     = 0x30  # Data register
REG_TEST     = 0x40  # Test register
REG_FILTER_L = 0x50  # Filter(low) register
REG_ZERO_CAL = 0x60  # Zero cal register 
REG_FULL_CAL = 0x70  # Full cal register
WRITE        = 0x00  # Write register
READ         = 0x08  # Read register
CH_1         = 0x00  # AIN1+ AIN1-
CH_2         = 0x01  # AIN2+ AIN2-
CH_3         = 0x02  # AIN1- AIN1-
CH_4         = 0x03  # AIN1- AIN2-
CH_MAX       = 0x04  # Cannot write to device register.

# Setting register
MD_NORMAL   = (0<<6)  # Normal
MD_CAL_SELF = (1<<6)  # Self calibrate 
MD_CAL_ZERO = (2<<6)  # Calibrate Zero
MD_CAL_FULL = (3<<6)  # Calibrate Full 
GAIN_1      = (0<<3)  # Gain x1 
GAIN_2      = (1<<3)  # Gain x2 
GAIN_4      = (2<<3)  # Gain x4
GAIN_8      = (3<<3)  # Gain x8
GAIN_16     = (4<<3)  # Gain x16
GAIN_32     = (5<<3)  # Gain x32
GAIN_64     = (6<<3)  # Gain x64
GAIN_128    = (7<<3)  # Gain x128
BIPOLAR     = (0<<2)  # Bipolar input
UNIPOLAR    = (1<<2)  # Unipolar input
BUF_NO      = (0<<1)  # Non-buffered input
BUF_EN      = (1<<1)  # Buffered input
FSYNC_0     = 0
FSYNC_1     = 1

# Filter low register
CLKDIS_0    = 0x00  # Clock enable
CLKDIS_1    = 0x10  # Clock disable (safe power)
CLK_4_9152M = 0x08  # 4.9152M
CLK_2_4576M = 0x00  # 2.4576M
CLK_1M      = 0x04  # 1M
CLK_2M      = 0x0C  # 2M
ZERO_0      = 0x00
ZERO_1      = 0x80
FS_50HZ     = 0x000
FS_60HZ     = 0x100
FS_250HZ    = 0x200
FS_500HZ    = 0x400
FS_DEFAULT  = 0xF00


class Tm7707():
    ref = 2.5
    channel = CH_1
    gain = GAIN_1
    polar = BIPOLAR
    buffered = BUF_NO
    clock_dis = CLKDIS_0
    clock_sel = CLK_4_9152M
    zero = ZERO_0
    fs = FS_500HZ

    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None, rst=None, drdy=None, init=True ):
        self.controller = controller
        self.controller.spiInit(sdi, sdo, sck, cs, cpol=True, cpha=True)
        self.rst = rst
        self.drdy = drdy
        self.controller.pinInput(drdy)
        self.controller.pinOutputHigh(rst)
        self.reset()
        if init:
            self.init()
   
    def reset( self ):
        self.controller.pinClr(self.rst)
        time.sleep(0.1)
        self.controller.pinSet(self.rst)
        time.sleep(0.1)

    def init( self ):
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_NORMAL | self.gain | self.polar | self.buffered ] )
        self.controller.spi( [ REG_FILTER_H | WRITE | self.channel,
                self.clock_dis | self.clock_sel | self.zero | ((self.fs>>8)&0xF) ] )
        self.controller.spi( [ REG_FILTER_L | WRITE | self.channel,
                self.fs&0xFF ] )
        # write default calib value (useless for following self-calibrating command)
        #self.controller.spi( [ REG_ZERO_CAL | WRITE | self.channel,
        #        0x1F, 0x40, 0x00 ] )
        #self.controller.spi( [ REG_FULL_CAL | WRITE | self.channel,
        #        0x57, 0x61, 0xAB ] )

    def readAllRegs( self ):
        val = {}
        _, a = self.controller.spiRead( [ REG_COMM | READ | self.channel, 0xFF ] )
        val['comm'] = a
        _, a = self.controller.spiRead( [ REG_SETUP | READ | self.channel, 0xFF ] )
        val['setup'] = a
        _, a = self.controller.spiRead( [ REG_FILTER_H | READ | self.channel, 0xFF ] )
        val['filter_h'] = a
        _, a = self.controller.spiRead( [ REG_FILTER_L | READ | self.channel, 0xFF ] )
        val['filter_l'] = a
        _, a, b, c = self.controller.spiRead( [ REG_ZERO_CAL | READ | self.channel, 0xFF, 0xFF, 0xFF ] )
        val['cal_zero'] = (a<<16)+(b<<8)+c
        _, a, b, c = self.controller.spiRead( [ REG_FULL_CAL | READ | self.channel, 0xFF, 0xFF, 0xFF ] )
        val['cal_full'] = (a<<16)+(b<<8)+c
        return val
   
    def sync( self ):
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_NORMAL | self.gain | self.polar | self.buffered | FSYNC_1 ] )
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_NORMAL | self.gain | self.polar | self.buffered | FSYNC_0 ] )
 
    def dataIsReady( self ):
        return self.controller.pinIsLow(self.drdy)

    def readData( self, mode=None ):
        _, a, b, c = self.controller.spiRead( [
                REG_DATA | READ | self.channel, 0xFF, 0xFF, 0xFF ] )
        val = (a<<16)|(b<<8)|c
        if mode in ['hex', 'int']:
            return val
        else:
            if self.polar == BIPOLAR:
                val -= 0x800000
                return self.ref * val / (2**23)
            else:
                return self.ref * val / (2**24)

    def calibSelf( self ):
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_CAL_SELF | self.gain | self.polar | self.buffered ] )
        while not self.dataIsReady():
            time.sleep(0.1)
        
    def calibZero( self ):
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_CAL_ZERO | self.gain | self.polar | self.buffered] )
        while not self.dataIsReady():
            time.sleep(0.1)
 
    def calibFull( self ):
        self.controller.spi( [ REG_SETUP | WRITE | self.channel,
                MD_CAL_FULL | self.gain | self.polar | self.buffered] )
        while not self.dataIsReady():
            time.sleep(0.1)
 



# coding: utf8
__doc__ = 'motion sensors'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils

MPU6050_REGS_BY_IDX = {
    1: 'AUX_VDDIO',
    25: 'SMPLRT_DIV',
    26: 'CONFIG',
    27: 'GYRO_CONFIG',
    28: 'ACCEL_CONFIG',
    29: 'FF_THR',
    30: 'FF_DUR',
    31: 'MOT_THR',
    32: 'MOT_DUR',
    33: 'ZRMOT_THR',
    34: 'ZRMOT_DUR',
    35: 'FIFO_EN',
    36: 'I2C_MST_CTRL',
    37: 'I2C_SLV0_ADDR',
    38: 'I2C_SLV0_REG',
    39: 'I2C_SLV0_CTRL',
    40: 'I2C_SLV1_ADDR',
    41: 'I2C_SLV1_REG',
    42: 'I2C_SLV1_CTRL',
    43: 'I2C_SLV2_ADDR',
    44: 'I2C_SLV2_REG',
    45: 'I2C_SLV2_CTRL',
    46: 'I2C_SLV3_ADDR',
    47: 'I2C_SLV3_REG',
    48: 'I2C_SLV3_CTRL',
    49: 'I2C_SLV4_ADDR',
    50: 'I2C_SLV4_REG',
    51: 'I2C_SLV4_DO',
    52: 'I2C_SLV4_CTRL',
    53: 'I2C_SLV4_DI',
    54: 'I2C_MST_STATUS',
    55: 'INT_PIN_CFG',
    56: 'INT_ENABLE',
    58: 'INT_STATUS',
    59: 'ACCEL_XOUT_H',
    60: 'ACCEL_XOUT_L',
    61: 'ACCEL_YOUT_H',
    62: 'ACCEL_YOUT_L',
    63: 'ACCEL_ZOUT_H',
    64: 'ACCEL_ZOUT_L',
    65: 'TEMP_OUT_H',
    66: 'TEMP_OUT_L',
    67: 'GYRO_XOUT_H',
    68: 'GYRO_XOUT_L',
    69: 'GYRO_YOUT_H',
    70: 'GYRO_YOUT_L',
    71: 'GYRO_ZOUT_H',
    72: 'GYRO_ZOUT_L',
    73: 'EXT_SENS_DATA_00',
    74: 'EXT_SENS_DATA_01',
    75: 'EXT_SENS_DATA_02',
    76: 'EXT_SENS_DATA_03',
    77: 'EXT_SENS_DATA_04',
    78: 'EXT_SENS_DATA_05',
    79: 'EXT_SENS_DATA_06',
    80: 'EXT_SENS_DATA_07',
    81: 'EXT_SENS_DATA_08',
    82: 'EXT_SENS_DATA_09',
    83: 'EXT_SENS_DATA_10',
    84: 'EXT_SENS_DATA_11',
    85: 'EXT_SENS_DATA_12',
    86: 'EXT_SENS_DATA_13',
    87: 'EXT_SENS_DATA_14',
    88: 'EXT_SENS_DATA_15',
    89: 'EXT_SENS_DATA_16',
    90: 'EXT_SENS_DATA_17',
    91: 'EXT_SENS_DATA_18',
    92: 'EXT_SENS_DATA_19',
    93: 'EXT_SENS_DATA_20',
    94: 'EXT_SENS_DATA_21',
    95: 'EXT_SENS_DATA_22',
    96: 'EXT_SENS_DATA_23',
    97: 'MOT_DETECT_STATUS',
    99: 'I2C_SLV0_DO',
    100: 'I2C_SLV1_DO',
    101: 'I2C_SLV2_DO',
    102: 'I2C_SLV3_DO',
    103: 'I2C_MST_DELAY_CTRL',
    104: 'SIGNAL_PATH_RESET',
    105: 'MOT_DETECT_CTRL',
    106: 'USER_CTRL',
    107: 'PWR_MGMT_1',
    108: 'PWR_MGMT_2',
    114: 'FIFO_COUNTH',
    115: 'FIFO_COUNTL',
    116: 'FIFO_R_W',
    117: 'WHO_AM_I',
    }

MPU6050_REGS_BY_NAME = {}
for k, v in MPU6050_REGS_BY_IDX.items():
    MPU6050_REGS_BY_NAME[v] = k

class Mpu6050():

    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x68, scl=scl, sda=sda )
        self.reset()

    def reset(self):
        self.writeByName( "PWR_MGMT_1", 0x80 )  # device reset
        while True:
            if not self.readByName( "PWR_MGMT_1" ) & 0x80:
                break
        self.writeByName( "PWR_MGMT_1", 0x00 )  # internal 8M CLK
        self.writeByName( "PWR_MGMT_2", 0x00 )  # no standby mode
        self.writeByName( "AUX_VDDIO", 0x80 )  # AUX_VDDIO=VDD
        self.writeByName( "SMPLRT_DIV", 0x07 )  # sample rate: 0x07(125Hz)
        self.writeByName( "CONFIG", 0x06 )  # filter: 0x06(5Hz)
        self.writeByName( "GYRO_CONFIG", 0x18 )  # gyro setting: 0x18(no self test, 2000deg/s)
        self.writeByName( "ACCEL_CONFIG", 0x01 )  # accel setting: 0x01(no self test, 2G, 5Hz)
        #self.writeByName( "PWR_MGMT_1", 0x01 )  # user X/PLL Clk 
 
    def readByIdx( self, idx ):
        return self.controller.i2c( [idx], 1 )[0]

    def readByName( self, name ):
        return self.controller.i2c( [MPU6050_REGS_BY_NAME[name]], 1 )[0]

    def writeByIdx( self, idx, val ):
        self.controller.i2c( [idx, val] )

    def writeByName( self, name, val ):
        self.controller.i2c( [MPU6050_REGS_BY_NAME[name], val] )

    def printAllReg( self ):
        #regs = self.controller.i2c( [0x0], 120 )
        #for i,name in MPU6050_REGS_BY_IDX.items():
        #    print( '[%d / 0x%02X] 0x%02X (%s)'% (i, i, regs[i], name) )
        for i,name in MPU6050_REGS_BY_IDX.items():
            print( '[%d / 0x%02X] 0x%02X (%s)'% (i, i, self.readByName(name), name) )

    def getTemperature( self ):
        a,b = self.controller.i2c( [0x41], 2 )
        return 35 + (Utils.s2h( chr(b) + chr(a) ) + 521)/340.0

    def getAccelX( self ):
        a,b = self.controller.i2c( [0x3B], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 9.8
        
    def getAccelY( self ):
        a,b = self.controller.i2c( [0x3D], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 9.8
        
    def getAccelZ( self ):
        a,b = self.controller.i2c( [0x3F], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 9.8
 
    def getGyroX( self ):
        a,b = self.controller.i2c( [0x43], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 16.4
        
    def getGyroY( self ):
        a,b = self.controller.i2c( [0x45], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 16.4
        
    def getGyroZ( self ):
        a,b = self.controller.i2c( [0x47], 2 )
        return Utils.s2h( chr(b) + chr(a) ) / 16384.0 * 16.4
 


     
# code, bandwidth, data rate
RATE_CODES = [
    ( 0x6, 3.125, 6.25 ),
    ( 0x7, 6.25, 12.5 ),
    ( 0x8, 12.5, 25.0 ),  # set
    ( 0x9, 25.0, 50.0 ),
    ( 0xA, 50.0, 100.0 ),  # reset default
    ( 0xB, 100.0, 200.0 ),
    ( 0xC, 200.0, 400.0 ),
    ( 0xD, 400.0, 800.0 ),
    ( 0xE, 800.0, 1600.0 ),
    ( 0xF, 1600.0, 3200.0 ) ]


class Adxl345():
    I2C_ADDR = 0x53
    MAX_RANGE = 16  # g

    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( self.I2C_ADDR, scl=scl, sda=sda )
        self.reset()

    def reset(self):
        self.controller.i2c( [0x31, 0x0B] )  # DATA_FORMAT: full resolution, max range
        self.controller.i2c( [0x2C, 0x08] )  # BW_RATE: 12.5Hz bandwidth, 25Hz output rate
        self.controller.i2c( [0x2D, 0x08] )  # POWER_CTRL: Measure 
        self.controller.i2c( [0x2E, 0x80] )  # INT_ENABLE: DATA_READY 
        self.controller.i2c( [0x1E, 0x00] )  # OFSX: X-axis offset
        self.controller.i2c( [0x1F, 0x00] )  # OFSY: Y-axis offset
        self.controller.i2c( [0x20, 0x00] )  # OFSZ: Z-axis offset

    def printAllReg( self ):
        print( self.controller.i2c( [0x0], 57 ) )

    def getAccelX( self ):
        a,b = self.controller.i2c( [0x32], 2 )
        return Utils.s2h( chr(a) + chr(b) ) / 25.0
 
    def getAccelY( self ):
        a,b = self.controller.i2c( [0x34], 2 )
        return Utils.s2h( chr(a) + chr(b) ) / 25.0
 
    def getAccelZ( self ):
        a,b = self.controller.i2c( [0x36], 2 )
        return Utils.s2h( chr(a) + chr(b) ) / 25.0
 

class Adxl313(Adxl345):
    I2C_ADDR = 0x1D
    MAX_RANGE = 4  # g





# coding:utf8
__doc__ = 'mcush controller'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
import os
import re
import sys
import time
import base64
import binascii
import logging
from . import Env
from . import Utils
from . import Instrument
from . import Register



class Mcush( Instrument.SerialInstrument ):
    '''Mcush core'''
    DEFAULT_NAME = 'MCUSH'
    DEFAULT_IDN = re.compile( 'mcush,([0-9]+\.[0-9]+.*)' )
    DEFAULT_MULTILINE_INPUT_LINE_LIMIT = 50
    DEFAULT_CMD_LINE_LIMIT = 128
    DEFAULT_CMD_ARGV_LIMIT = 20

    regs_by_name = {}
    regs_by_addr = {}

    def addReg( self, r ):
        '''add register'''
        self.regs_by_name[r.name] = r
        self.regs_by_addr[r.address] = r
        
    def setReg( self, regname, value ):
        '''set register'''
        r = self.regs_by_name[regname]
        w, b, fget, fset = Register.REGISTER_TYPE[r.reg_type]
        m = fset(value)
        self.writeMem( r.address, m )
    
    def getReg( self, regname ):
        '''get register'''
        r = self.regs_by_name[regname]
        w, b, fget, fset = Register.REGISTER_TYPE[r.reg_type]
        m = self.readMem( r.address, length=b )
        return fget(m)

    def getLedNumber(self):
        r = self.writeCommand( 'led -n' )
        return int(r[0])
    
    def led( self, idx, on=None, toggle=None ):
        '''led control'''
        if on is None and toggle is None:
            # read
            r = self.writeCommand( 'led -i%d'% (idx) )
            return bool(r[0].strip() == '1')
        else:
            # set 
            cmd = 'led -i%d'% (idx)
            if toggle is not None:
                cmd += ' -t'
            else:
                cmd += ' -s' if on else ' -c'
            self.writeCommand( cmd )

    def ledSetValue( self, idx, val ):
        if isinstance(idx, list):
            for led in idx:
                self.led( led, on=val )
        else:
            self.led( idx, on=val )

    def ledOn( self, idx ):
        if isinstance(idx, list):
            for led in idx:
                self.led( led, on=True )
        else:
            self.led( idx, on=True )

    def ledOff( self, idx ):
        if isinstance(idx, list):
            for led in idx:
                self.led( led, on=False )
        else:
            self.led( idx, on=False )

    def ledToggle( self, idx ):
        if isinstance(idx, list):
            for led in idx:
                self.led( led, toggle=True )
        else:
            self.led( idx, toggle=True )

    ledNumber = getLedNumber
    ledNum = getLedNumber
    ledSetVal = ledSetValue
    ledSet = ledOn
    ledClr = ledOff
    ledReset = ledOff
    ledClear = ledOff

    # raw gpio command
    def gpio( self, port, i=None, o=None, s=None, c=None, t=None ):
        '''gpio control'''
        cmd = 'gpio -p%s'% str(port)
        if i is not None:
            if type(i) is int:
                cmd += ' -i0x%x'% i
            elif type(i) is bool and i:
                cmd += ' -i'
        if o is not None:
            if type(o) is int:
                cmd += ' -o0x%x'% o
            elif type(o) is bool and o:
                cmd += ' -o'
        if s is not None:
            if type(s) is int:
                cmd += ' -s0x%x'% s
            elif type(s) is bool and s:
                cmd += ' -s'
        if c is not None:
            if type(c) is int:
                cmd += ' -c0x%x'% c 
            elif type(c) is bool and c:
                cmd += ' -c'
        if t is not None:
            if type(t) is int:
                cmd += ' -t0x%x'% t 
            elif type(t) is bool and t:
                cmd += ' -t'
        ret = self.writeCommand( cmd )
        if ret:
            return eval(ret[0])

    # some partial functions for gpio manipulating
    def pinInput( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, i=True )
        else:
            self.gpio( pin, i=True )

    def pinOutput( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, o=True )
        else:
            self.gpio( pin, o=True )

    def pinOutputHigh( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, o=True, s=True )
        else:
            self.gpio( pin, o=True, s=True )

    def pinOutputLow( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, o=True, c=True )
        else:
            self.gpio( pin, o=True, c=True )

    def pinSetVal( self, pin, val ):
        if isinstance(pin, list):
            for p in pin:
                if val:
                    self.gpio( p, s=True )
                else:
                    self.gpio( p, c=True )
        else:
            if val:
                self.gpio( pin, s=True )
            else:
                self.gpio( pin, c=True )

    pinSetValue = pinSetVal

    def pinSet( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, s=True )
        else:
            self.gpio( pin, s=True )

    pinSetHigh = pinSet

    def pinClr( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, c=True )
        else:
            self.gpio( pin, c=True )

    pinSetLow = pinClr
    pinReset = pinClr

    def pinToggle( self, pin ):
        if isinstance(pin, list):
            for p in pin:
                self.gpio( p, t=True )
        else:
            self.gpio( pin, t=True )

    def pinRead( self, pin ):
        if isinstance(pin, list):
            return [self.gpio(p) for p in pin]
        else:
            return self.gpio( pin )

    def pinIsHigh( self, pin ):
        return bool(self.gpio( pin ))

    pinIsSet = pinIsHigh

    def pinIsLow( self, pin ):
        return not bool(self.gpio( pin ))

    pinIsClear = pinIsLow
    pinIsReset = pinIsLow



    def parseMemLine( self, line, compact_mode=False ):
        '''parse memory line which has been stripped'''
        # format(standard):  XXXXXXXX: xx xx xx xx xx ... xx    (no ascii mode, final space stripped)
        # format(compact):   xxxxxxxxxxxx...xx
        #print( type(line), line )
        line_len = len(line)
        line_len_err = bool(line_len % 2) if compact_mode else bool((line_len-9) % 3)
        if line_len_err:
            raise CommandExecuteError('memory line width error, length=%s'% line_len)
        if compact_mode:
            m = [binascii.unhexlify(line[i:i+2]) for i in range(0,line_len,2)]
        else:
            m = [binascii.unhexlify(line[i:i+2]) for i in range(10,line_len,3)]
        ret = Env.EMPTY_BYTE.join(m)
        #print( type(ret), len(ret), ret )
        return ret
       
    def fillMem( self, addr, pattern, width, length ):
        '''fill memory with specific pattern'''
        cmd = 'mfill -b 0x%X -w %d -p 0x%X -l %d'% ( addr, width, pattern, length )
        self.writeCommand( cmd )
 
    def testMem( self, addr, pattern, width, length ):
        '''test filled memory with specific pattern'''
        cmd = 'mfill -t -b 0x%X -w %d -p 0x%X -l %d'% ( addr, width, pattern, length )
        try:
            self.writeCommand( cmd )
            return True
        except Instrument.CommandExecuteError:
            return False
        
    def malloc( self, length ):
        cmd = 'mapi -m -l %d'% length
        ret = self.writeCommand(cmd)
        addr = eval(ret[0])
        return addr

    def realloc( self, addr, length ):
        cmd = 'mapi -r -b 0x%08X -l %d'% (addr, length)
        ret = self.writeCommand(cmd)
        addr = eval(ret[0])
        return addr

    def free( self, addr ):
        cmd = 'mapi -f -b 0x%08X'% addr
        self.writeCommand( cmd )

    def mkbuf( self, value_list, float_mode=False ):
        cmd = 'mkbuf -f' if float_mode else 'mkbuf'
        self.setPrompts( self.DEFAULT_PROMPTS_MULTILINE )
        self.writeCommand( cmd )
        line = ''
        for v in value_list:
            if float_mode:
                line += '%f '% v
            else:
                line += '%d '% v
            if len(line) > self.DEFAULT_MULTILINE_INPUT_LINE_LIMIT:
                self.writeCommand( line.rstrip() )
                line = ''
        if line:
            self.writeCommand( line.rstrip() )
        self.setPrompts()
        r = Utils.parseKeyValueLines(self.writeCommand(''))
        return int(r['address'], 16)
            
 
    def readMem( self, addr, length=1, compact_mode=False, retry=None ):
        '''get memory'''
        cmd = 'x -b 0x%X -l %d'% ( addr, length )
        if compact_mode:
            cmd += ' -c'
        if retry:
            ret = self.writeCommandRetry( cmd, retry )
        else:
            ret = self.writeCommand( cmd )
        assert 1 <= len(ret)
        for line in ret:
            self.logger.info( line )
        mem = Env.EMPTY_BYTE.join( \
               [self.parseMemLine(line.strip(), compact_mode) for line in ret])
        #print( type(mem), len(mem), mem )
        return mem[:length]

    MAX_ITEMS_PER_WRITE = { 1: 16, 2: 8, 4: 4  }

    def writeMem( self, addr, data, width=1 ):
        '''write memory'''
        length = len(data)
        written = 0
        while written < length:
            cmd = 'w -b 0x%X -w%d '% ( addr, width )
            left = min(length-written, self.MAX_ITEMS_PER_WRITE[width])
            if width == 1:
                cmd += ' '.join([str(ord(c)) for c in data[written:written+left]])
            elif width == 2:
                cmd += ' '.join([hex(ord(data[i])+(ord(data[i+1])<<8)) for i in range(written, written+left*2, 2)])
            elif width == 4:
                cmd += ' '.join([hex(ord(data[i])+(ord(data[i+1])<<8)+(ord(data[i+2])<<16)+(ord(data[i+3])<<24)) for i in range(written, written+left*4, 4)])
            self.writeCommand( cmd )
            written += left*width
            addr += left*width

    def dumpMem( self, addr, length=1 ):
        '''dump memory'''
        Utils.dumpMem( self.readMem(addr, length) )

    def uptime( self ):
        return self.writeCommand('uptime')[0].strip()

    def sysTask( self ):
        ret = self.writeCommand('sys t')
        return ret

    def sysQueue( self ):
        ret = self.writeCommand('sys q')
        return ret

    def errnoStop( self ):
        '''stop errno blink'''
        self.writeCommand( 'error -s' )
        
    def errno( self, new=None, stop=False ):
        '''set/get error number'''
        if stop:
            self.writeCommand( 'error -s' )
            return 
        if new is None: 
            ret = self.writeCommand( 'error' )
            return int(ret[0].strip())
        else:
            assert isinstance(new, int)
            self.writeCommand( 'error %d'% new )

    def beep( self, freq=None, duration=0.05, times=1 ):
        cmd = 'beep'
        if freq:
            cmd += ' -f%d'% freq
        if duration:
            cmd += ' %d'% (int(duration*1000))
        while times:
            self.writeCommand(cmd)
            if times <= 1:
                break
            times -= 1
            time.sleep( duration )
    
    def disp( self, integer=None, buf=None ):
        if integer is not None:
            cmd = 'disp -i %d'% integer
        elif buf is not None:
            cmd = 'disp -- "%s"'% buf
        self.writeCommand(cmd)

    def list( self, pathname=None ):
        if pathname:
            cmd = 'ls %s'% pathname
        else:
            cmd = 'ls'
        ret = self.writeCommand(cmd)
        flist = []
        path = ''
        for l in ret:
            if l.startswith('/'):
                path = l.strip().rstrip(':')
                flist.append( (path, None, None) )
            else:
                a, b = l.lstrip().split('  ')
                flist.append( (path, b, int(a)) )
        return flist

    def checkFileExist( self, pathname ):
        exist = False
        for p, f, s in self.list( pathname ):
            if s is not None:
                exist = True
        return exist

    def checkFileSize( self, pathname ):
        for p, f, s in self.list( pathname ):
            if s is not None:
                return s
        return None

    def convPathname( self, pathname ):
        if not Env.PYTHON_V3:
            pathname = unicode(pathname)
        else:
            pathname = str(pathname)
        namelen = len(os.path.basename(pathname).encode('utf8'))
        if namelen > 28:
            raise Exception( "File name length too long %d"% namelen )
        if pathname.find(' ') != -1:
            pathname = '\"' + pathname + '\"'
        return pathname

    def cat( self, pathname, b64=False, write=False, append=False, buf='' ):
        pathname = self.convPathname(pathname)
        if write or append:
            if append:
                cmd = 'cat -a '
            else:
                cmd = 'cat -w '
            if b64:
                cmd += '-b '
            cmd += pathname
            self.setPrompts( self.DEFAULT_PROMPTS_MULTILINE )
            self.writeCommand( cmd )
            if b64:
                buf = base64.encodestring(buf)
            buf = buf.splitlines()
            for l in buf:
                self.writeCommand( l.rstrip() )
            self.setPrompts()
            self.writeCommand( '' )
        else:
            if b64:
                cmd = 'cat -b '
            else:
                cmd = 'cat '
            cmd += pathname
            ret = '\n'.join(self.writeCommand( cmd ))
            if b64:
                if Env.PYTHON_V3 and isinstance(ret, str):
                     ret = ret.encode('utf8')
                ret = base64.decodestring( ret )
            return ret

    def writeFile( self, pathname, buf='' ):
        self.cat( pathname, b64=True, write=True, append=False, buf=buf )

    def appendFile( self, pathname, buf='' ):
        self.cat( pathname, b64=True, write=True, append=True, buf=buf )

    def remove( self, pathname ):
        pathname = self.convPathname(pathname)
        cmd = 'rm ' + pathname
        self.writeCommand( cmd )

    def rename( self, old_pathname, new_name ):
        old_pathname = self.convPathname(old_pathname)
        new_name = self.convPathname(new_name)
        cmd = 'rename %s %s'% (old_pathname, new_name)
        self.writeCommand( cmd )

    def getFile( self, pathname, local_pathname ):
        # read remote_file and save locally
        dat = self.cat( pathname, b64=True )
        open( local_pathname, 'wb+' ).write(dat)

    def putFile( self, pathname, local_pathname, segment_size=512, segment_done_callback=None ):
        # split local_file into pieces and write
        dat = open(local_pathname, 'rb').read()
        dat_size = len(dat)
        if dat_size == 0:
            return
        dat_segments = int((dat_size-1) / segment_size) + 1
        # write the first
        d = dat[0:segment_size]
        self.cat( pathname, b64=True, write=True, buf=d )
        sent = len(d)
        if segment_done_callback:
            segment_done_callback(1, dat_segments, sent, dat_size)
        # append the remainings
        for i in range(1, dat_segments):
            d = dat[segment_size*i:segment_size*(i+1)]
            self.cat( pathname, b64=True, write=True, append=True, buf=d )
            sent += len(d) 
            if segment_done_callback:
                segment_done_callback(1+i, dat_segments, sent, dat_size)
 
    def spiffs( self, command, value=None, addr=None, compact_mode=None ):
        cmd = 'spiffs -c %s'% command
        if value is not None:
            cmd += ' %s'% value
        if addr is not None:
            cmd += ' -b 0x%X'% addr
        if compact_mode:
            cmd += ' --compact'
        return self.writeCommand( cmd )

    def spiffsID( self ):
        return int(self.spiffs('id')[0], 16)
    
    def spiffsStatus( self ):
        return int(self.spiffs('status')[0], 16)

    def spiffsMount( self ):
        self.spiffs('mount')

    def spiffsUmount( self ):
        self.spiffs('umount')

    def spiffsRemount( self ):
        self.spiffs('remount')

    def spiffsFormat( self ):
        oldtimeout = self.setTimeout( 60 )
        self.spiffs( 'format' )
        self.setTimeout( oldtimeout )

    def spiffsErase( self, addr=None ):
        oldtimeout = self.setTimeout( 60 )
        if addr is not None:
            self.spiffs( 'erase', addr=addr )
        else:
            self.spiffs( 'erase' )
        self.setTimeout( oldtimeout )

    def spiffsInfo( self ):
        var = Utils.parseKeyValueLines( self.spiffs( 'info' ) )
        var['total'] = int(var['total'])
        var['used'] = int(var['used'])
        var['free'] = var['total'] - var['used']
        return var

    def spiffsReadPage( self, page, pagesize=256 ):
        compact_mode = True
        ret = self.spiffs( "read", addr=page*pagesize, compact_mode=compact_mode ) 
        mem = Env.EMPTY_BYTE.join( \
               [self.parseMemLine(line.strip(), compact_mode=compact_mode) for line in ret])
        return mem
 
    def spiffsWritePage( self, page, buf, pagesize=256 ):
        self.setPrompts( self.DEFAULT_PROMPTS_MULTILINE )
        ret = self.spiffs( "write", addr=page*pagesize ) 
        line = ''
        for item in buf:
            if isinstance(item, int):
                line += '%d '% item
            elif isinstance(item, str):
                line += '%d '% ord(item)
            if len(line) > self.DEFAULT_MULTILINE_INPUT_LINE_LIMIT:
                self.writeCommand( line.rstrip() )
                line = ''
        if line:
            self.writeCommand( line.rstrip() )
        self.setPrompts()
        self.writeCommand( '' )
    
    def fcfsFormat( self ):
        self.writeCommand( 'fcfs -c format' )
 
    def fcfsProgram( self, mem ):
        # append zero to be 4-bytes alignment
        while len(mem) & 0x3:
            mem += '\x00'
        # prepare 32bit word list
        words = [Utils.s2I(mem[4*i:4*(i+1)]) for i in range(len(mem)/4)]
        offset = 0
        while words:
            cmd = 'fcfs -c program'
            if offset:
                cmd += ' -o %d'% (offset*4)
            count_argv = len(cmd.split(' '))
            while words:
                wr = ' %d'% words[0] 
                if len(cmd + wr) > self.DEFAULT_CMD_LINE_LIMIT:
                    break
                words.pop(0)
                cmd += wr
                offset += 1
                count_argv += 1
                if count_argv >= self.DEFAULT_CMD_ARGV_LIMIT:
                    break
            self.writeCommand( cmd )

    def sgpioStop( self ):
        self.writeCommand( 'sgpio --stop' )

    def sgpioStart( self ):
        self.writeCommand( 'sgpio --start' )

    def sgpioOutput( self, port, pins, buf, freq=10, loop=True, start=True ):
        self.sgpioStop()
        cmd = 'sgpio -p%s -o0x%X -f%d'% (int(port), int(pins), int(freq))
        if loop:
            cmd += ' -l'
        self.setPrompts( self.DEFAULT_PROMPTS_MULTILINE )
        self.writeCommand( cmd )
        line = ''
        for item in buf:
            line += '%d '% item
            if len(line) > self.DEFAULT_MULTILINE_INPUT_LINE_LIMIT:
                self.writeCommand( line.rstrip() )
                line = ''
        if line:
            self.writeCommand( line.rstrip() )
        self.setPrompts()
        self.writeCommand( '' )
        if start:
            self.sgpioStart()

    def sgpioInput( self, port, pins, freq, length ):
        self.sgpioStop()
        cmd = 'sgpio -p%s -i0x%X -f%d --input_len=%d'% (int(port), int(pins), int(freq), int(length))
        self.writeCommand( cmd )
        self.sgpioStart()
        while True:
            r = self.writeCommand( 'sgpio --info' )
            if r[0].startswith("stop"):
                addr = int(r[1].split('  ')[1].split(':')[-1], 16)
                return addr
            time.sleep(0.1)
 
    # emulated i2c bus control
    def i2c_init( self, addr, scl=None, sda=None, lsb=None, delay=None ):
        cmd = 'i2c -a0x%X -I' % (addr)
        if scl:
            cmd += ' --scl=%s'% scl
        if sda:
            cmd += ' --sda=%s'% sda
        if lsb:
            cmd += ' --lsb'
        if delay:
            cmd += ' --delay=%s'% delay
        self.writeCommand( cmd )

    i2cInit = i2c_init

    def i2c( self, write=[], read=None, addr=None, no_stop_bit=None ):
        cmd = 'i2c'
        if addr is not None:
            cmd += ' -a0x%X'% addr
        if read:
            cmd += ' -r%d'% read
        if no_stop_bit:
            cmd += ' -n'
        if write:
            cmd += ' %s'% (' '.join([str(i) for i in write])) 
        ret = self.writeCommand( cmd )
        if read:
            r = []
            for l in ret:
                for v in l.strip().split():
                    r.append(eval(v))
            return r
        else:
            return None

    # emulated spi bus control
    def spi_init( self, sdi=None, sdo=None, sck=None, cs=None, cpol=None, cpha=None, width=None, lsb=None, delay=None ):
        cmd = 'spi -I'
        if sdi:
            cmd += ' --sdi=%s'% sdi
        if sdo:
            cmd += ' --sdo=%s'% sdo
        if sck:
            cmd += ' --sck=%s'% sck
        if cs:
            cmd += ' --cs=%s'% cs
        if cpol:
            cmd += ' --cpol'
        if cpha:
            cmd += ' --cpha'
        if lsb:
            cmd += ' --lsb'
        if width:
            cmd += ' --width=%d'% width
        if delay:
            cmd += ' --delay=%s'% delay
        self.writeCommand( cmd )
    
    def spi_update( self, sdi=None, sdo=None, sck=None, cs=None, cpol=None, cpha=None, width=None, lsb=None, delay=None ):
        cmd = 'spi -U'
        if sdi:
            cmd += ' --sdi=%s'% sdi
        if sdo:
            cmd += ' --sdo=%s'% sdo
        if sck:
            cmd += ' --sck=%s'% sck
        if cs:
            cmd += ' --cs=%s'% cs
        if cpol:
            cmd += ' --cpol'
        if cpha:
            cmd += ' --cpha'
        if lsb:
            cmd += ' --lsb'
        if width:
            cmd += ' --width=%d'% width
        if delay:
            cmd += ' --delay=%s'% delay
        self.writeCommand( cmd )
 
    def spi_deinit( self ):
        self.writeCommand( 'spi -D' )
 
    def spi( self, write=[], read=None ):
        cmd = 'spi'
        if read:
            cmd += ' -r'
        for d in write:
            cmd += ' %d'% d 
        ret = self.writeCommand( cmd )
        if read:
            r = []
            for l in ret:
                for v in l.strip().split():
                    r.append(eval(v))
            return r
        else:
            return None
    
    def spi_write( self, write=[] ):
        if isinstance(write, list):
            return self.spi( write )
        else:
            return self.spi( [write] )

    def spi_read( self, write=[] ):
        if isinstance(write, list):
            return self.spi( write, read=True )
        else:
            return self.spi( [write], read=True )

    spiInit = spi_init
    spiUpdate = spi_update
    spiDeinit = spi_deinit
    spiWrite = spi_write
    spiRead = spi_read

    # emulated pulse generator
    def pulse_init( self, pin='0.0', delay_us=None, invert=None ):
        cmd = 'pulse -p%s -I'% pin
        if delay_us is not None:
            cmd += ' --delay=%d'% delay_us
        if invert:
            cmd += ' --invert'
        self.writeCommand( cmd )

    def pulse_deinit( self ):
        self.writeCommand( 'pulse -D' )
        
    def pulse( self, number ):
        self.writeCommand( 'pulse %d'% number )

    pulseInit = pulse_init
    pulseDeinit = pulse_deinit

    def pwm_init( self, freq=None, range_value=None, init_value=None ):
        cmd = 'pwm -I'
        if freq is not None:
            cmd += ' -f%s'% int(freq)
        if range_value is not None:
            cmd += ' -r%s'% int(range_value)
        if init_value is not None:
            cmd += ' -v%s'% int(init_value)
        self.writeCommand( cmd )

    def pwm_deinit( self ):
        cmd = 'pwm -D'
        self.writeCommand( cmd )
        
    pwmInit = pwm_init
    pwmDeinit = pwm_deinit

    # hal layer pwm control
    def pwm( self, index=None, value=50 ):
        cmd = 'pwm'
        if index is not None:
            cmd += ' -i%d'% index
        cmd += ' -v%d'% value
        self.writeCommand( cmd )

    def adc_init( self ):
        self.writeCommand( "adc -I" )
    
    def adc_deinit( self ):
        self.writeCommand( "adc -D" )
     
    adcInit = adc_init
    adcDeinit = adc_deinit

    def adc( self, channel=None ):
        if channel is None:
            ret = self.writeCommand( "adc" )
            return [float(v) for v in ret[0].split(',')]
        else:
            ret = self.writeCommand( "adc -i%d"% channel )
            return float(ret[0])

    def rtcSync( self ):
        # update to current time
        t = time.localtime()
        cmd = 'rtc -s %d-%d-%d %d:%d:%d'% (t.tm_year, t.tm_mon, t.tm_mday,
                                           t.tm_hour, t.tm_min, t.tm_sec )
        self.writeCommand( cmd )
   
    def rtcRead( self ): 
        try:
            r = self.writeCommand( 'rtc' )
            d, t = r[0].split(' ')
        except Instrument.CommandExecuteError:
            return None
        YEAR, MONTH, MDAY = d.split('-')
        HOUR, MIN, SEC = t.split(':')
        return (int(YEAR), int(MONTH), int(MDAY), int(HOUR), int(MIN), int(SEC), 0, 1, -1)

    def log( self, message=None, enable=None ):
        cmdline = 'log'
        if enable is not None:
            cmdline += ' -e' if enable else ' -d'
        if message is not None:
            cmdline += ' -m \"%s\"'% str(message)
        return self.writeCommand( cmdline )

    def logTail( self ):
        cmd = 'log -t'
        r = self.writeCommand( cmd )
        return '\n'.join(r[1:]) + '\n'

    def logEnter( self ):
        self.writeLine('log')
        self.port.read(4)

    def logExit( self ):
        self.port.write( self.DEFAULT_TERMINATOR_RESET )
        self.port.flush()
        self.readUntilPrompts()

    def logReadline( self ):
        return self.readLine()

    def ping( self, host, times=10, silent=True ):
        cmd = 'ping %s'% host
        self.writeLine(cmd)
        self.port.read(len(cmd)+1) 
        send, recv, ms = 0, 0, 0
        while True:
            line = self.readLine()
            if ('failed' in line) or ('timeout' in line):
                self.readUntilPrompts()
                return (0, 0, 0)
            if 'ping: send' in line:
                send += 1
                if send > times:
                    break
            elif 'ping: recv' in line:
                recv += 1
                # parse response line in format:
                # example: ping: recv 10.168.1.1  8 ms
                ms += int(line.split(' ')[-2])
                if send >= times:
                    break
            if not silent:
                print( line )
        if recv:
            ms /= recv
        self.port.write( self.DEFAULT_TERMINATOR_RESET )
        self.port.flush()
        self.readUntilPrompts()
        return (send, recv, ms)

    def netstat( self ):
        r = self.writeCommand( "netstat" )
        return Utils.parseKeyValueLines(r)

    def setNewIP( self, ip, netmask, gateway ):
        self.setPrompts( self.DEFAULT_PROMPTS_MULTILINE )
        self.writeCommand('netstat -c ip')
        self.writeCommand( str(ip) )
        self.writeCommand( str(netmask) )
        self.writeCommand( str(gateway) )
        self.setPrompts()
        self.writeCommand('')
        

    DEFAULT_TIMEOUT_UPRADE = 20
    def upgrade( self, upgrade_file='/s/upgrade.bin' ):
        command = 'upgrade -f %s'% upgrade_file
        try:
            self.setTimeout( self.DEFAULT_TIMEOUT_UPRADE )
            self.writeCommand( command )
        except Instrument.CommandTimeoutError:
            pass

    def wget( self, url, local_file, timeout=30 ):
        command = 'wget -u %s -f %s'% (url, local_file)
        oldtimeout = self.setTimeout( timeout )
        try:
            ret = self.writeCommand( command )
        except Instrument.CommandTimeoutError:
            self.port.write( self.DEFAULT_TERMINATOR_RESET )
            self.port.flush()
            self.readUntilPrompts()
            ret = None
        self.setTimeout( oldtimeout )
        if ret is None:
            return False
        if len(ret):
            if 'bytes saved' in ret[-1]:
                return True
        return False 

    def crc( self, pathname ):
        command = 'crc %s'% pathname 
        ret = self.writeCommand( command )
        return int(ret[0], 16)
        
 
    def luaReset( self ):
        self.writeCommand( 'lua -r' )

    def luaStop( self ):
        self.writeCommand( 'lua -s' )
 
    def luaCommand( self, cmd ):
        ret = self.writeCommand( 'lua -c "%s"'% cmd )
        return ret[:-1]

    def can( self, cmd, index=None, value=None, ext=None, rtr=None, args=None ):
        command = 'can -c %s'% cmd
        if index is not None:
            command += ' -i %d'% index
        if value is not None:
            command += ' -v %d'% value
        if ext:
            command += ' -e'
        if rtr:
            command += ' -r'
        if args:
            command += ' '+(' '.join([str(a) for a in args]))
        ret = self.writeCommand( command )
        return ret

    def canBaudrate( self, new_baudrate ):
        self.can( 'baudrate', value=new_baudrate )

    def canWrite( self, cob_id, dat=[], ext=None, rtr=None ):
        #self.logger.debug( 'canWrite id=0x%X, ext=%d, rtr=%d, dat=%s'% (cob_id, int(bool(ext)), int(bool(rtr)), ','.join([hex(d) for d in dat])) )
        self.can( 'write', index=cob_id, ext=ext, rtr=rtr, args=dat )

    def canReset( self ):
        self.can( 'reset' )

    def canResetInput( self ):
        self.can( 'reset_input' )

    def canRead( self, lines=None, block_ms=None  ):
        ret = []
        for line in self.can( 'read', index=lines, value=block_ms ):
            items = line.split()
            cid, rtr = items[0], items[1]
            try:
                dat = items[2]
            except IndexError:
                dat = ''
            id = int(cid,16)
            ext = bool(len(cid)>3)
            rtr = bool(rtr=='R')
            #self.logger.debug( 'canRead id=0x%X, ext=%d, rtr=%d, dat=%s'% (id, int(ext), int(rtr), dat) )
            ret.append( (id, ext, rtr, binascii.unhexlify(dat)) )
        return ret

    def env( self ):
        return Utils.parseKeyValueLines( self.writeCommand('env') )




# coding:utf8
__doc__ = 'beep morse code with beep command (or customized function)'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
# refer to http://en.wikipedia.org/wiki/Morse_code
import sys
import time
from .. import Mcush


MorseCode = {
'a' : '.-', 
'b' : '-...', 
'c' : '-.-.', 
'd' : '-..', 
'e' :'.',
'f' : '..-.', 
'g' : '--.', 
'h' : '....', 
'i' : '..', 
'j' : '.---',
'k' : '-.-', 
'l' : '.-..', 
'm' : '--', 
'n' : '-.', 
'o' : '---',
'p' : '.--.', 
'q' : '--.-', 
'r' : '.-.', 
's' : '...', 
't' : '-',
'u' : '..-', 
'v' : '...-', 
'w' : '.--', 
'x' : '-..-', 
'y' : '-.--',
'z' : '--..', 
'0' : '-----', 
'1' : '.----', 
'2' : '..---', 
'3' : '...--',
'4' : '....-', 
'5' : '.....', 
'6' : '-....', 
'7' : '--...', 
'8' : '---..',
'9' : '----.', 
'/' : '-..-.', 
'+' : '.-.-.', 
'=' : '-...-', 
'.' : '.-.-.-',
',' : '--..--', 
'?' : '..--..', 
'(' : '-.--.', 
')' : '-.--.-', 
'-' : '-....-',
'\"' : '.-..-.', 
'_' : '..--.-', 
'\'' : '.----.', 
':' : '---...', 
';' : '-.-.-.', 
'$' : '...-..-', 
'&' : '.-...', 
'@' : '.--.-.', 
} 


class MorseCodeBeeper():

    freq = 500
    duration_dot = 0.1
    duration_dash = 0.1 * 3  # equal to 3 dot
    pause_bit = 0.1  # equal to 1 dot
    pause_char = 0.1 * 3  # equal to 3 dots
    pause_word = 0.1 * 7  # equal to 7 dots

    beeper = None

    def __init__( self, controller ):
        self.controller = controller

    def beeper_led( self, index, duration ):
        self.controller.led( index, True )
        time.sleep( duration )
        self.controller.led( index, False )

    def beeper_led0( self, duration ):
        self.beeper_led( 0, duration )

    def beeper_led1( self, duration ):
        self.beeper_led( 1, duration )

    def beeper_beep( self, duration ):
        self.controller.beep( duration=duration )
        time.sleep( duration )

    def beep_function( self, duration ):
        if self.beeper:
            self.beeper( duration )
        else:
            self.beeper_beep( duration )

    def set_beeper( self, func ):
        self.beeper = func

    def BeepChar( self, char ):
        if not char in MorseCode:
            return
        #print '%c'%char,
        for bit in list(MorseCode[char]):
            duration = self.duration_dot if bit == '.' else self.duration_dash
            self.beep_function(duration)
            time.sleep(self.pause_bit)

    def BeepWord( self, word ):
        '''Beep word'''
        word = word.lower().strip()
        for i in range(len(word)):
            #print word[i],
            sys.stdout.write( word[i] )
            sys.stdout.flush()
            self.BeepChar( word[i] )
            time.sleep(self.pause_char)
            
    def BeepString( self, string ):
        '''Beep ASCII string'''
        for word in string.strip().split(' '):
            word = word.strip()
            if word:
                #print word
                self.BeepWord(word)
                time.sleep(self.pause_word)

    def BeepFile( self, filename ):
        '''Read and beep ASCII file'''
        for string in open(filename,'r').readlines():
            string = string.strip()
            if string:
                self.BeepString( string )


    beepString = BeepString
    beepStr = BeepString
    beepChar = BeepChar
    beepWord = BeepWord
    beepFile = BeepFile




# coding:utf8
__doc__ = 'products from Shanghai Yaohua Weighing System'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env

FUNCTIONS = {
'w': 'Weight',
'n': 'Net-weight',
't': 'Tare-weight'
}

PRECESION = [1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]

FORMATTER = {
1.0: "%.0f",
0.1: "%.1f",
0.01: "%.2f",
0.001: "%.3f",
0.0001: "%.4f",
0.00001: "%.5f",
}


# serial data demo:
# ww000.000lb
# wn-0000.1kg  (negative)
# wt000.000kg

class XK3190A12( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'XK3190'
    DEFAULT_PACKET_FORMAT = re_compile( 'w[wnt]-?[0-9]*.[0-9]*(kg|lb)' )

    def __init__( self, *args, **kwargs ):
        kwargs['check_idn'] = False  # not support scpi commands
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 

    def connect( self, check_idn=False ):
        self.port.connect()
        if not self.port.connected:
            return

    def writeCommand( self, cmd ):
        # only receives serial packets
        if cmd not in ['R', 'T', 'Z']:
            raise Exception('Invalid command')
        self.port.write(cmd)
        self.port.flush()

    def formatReading( self, reading, precision ):
        return FORMATTER[precision]% reading

    def getReading( self ):
        # read for one packet and parse result
        while True:
            line = self.readLine( eol='\x0A', timeout=self.DEFAULT_TIMEOUT ).strip()
            if not self.DEFAULT_PACKET_FORMAT.match( line ):
                continue
            # parse packet
            function = line[1]
            if function in FUNCTIONS:
                function = FUNCTIONS[function]
            unit = line[-2:]
            reading_str = line[2:-2]
            try:
                precesion = PRECESION[len(reading_str.split('.')[-1])]
            except IndexError:
                precesion = PRECESION[-1]
            reading = float(reading_str)
            return function, reading, precesion, unit





# coding: utf8
__doc__ = 'ws2812 color led strap controller module'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush
from . import Bitmap
from . import Font



class LEDS():
    '''WS2812 controller'''
    
    def __init__( self, controller, length=None, group_length=None, swap_rg=None, pin=None ):
        if length is None:
            if group_length is None:
                length = self.width * self.height
            else:
                length = 1  # whole strap share the same color
        self.length = length
        self.swap_rg = swap_rg
        self.controller = controller
        cmd = "W -I -l%d"% length
        if group_length:
            cmd += " -G%d"% group_length
        if pin is not None:
            cmd += " -p%s"% pin
        self.controller.writeCommand( cmd )

    DEFAULT_CMD_LINE_LIMIT = 110

    def write( self, mem, offset=0, push=None ):
        mem = mem[:self.length-offset]  # ignore the invalid portion
        cmd = 'W'
        if offset:
            cmd += ' -o%d'% offset
        if self.swap_rg:
            cmd += ' -g'
        if push is not None:
            if push == 'f':
                cmd += ' -f'
            elif push == 'b':
                cmd += ' -b'
        count_data = 0
        count_argv = len(cmd.split())
        while True:
            if len(mem) == 0:
                if count_data:
                    # the final line, add -w option
                    self.controller.writeCommand( cmd.replace('W', 'W -w') )
                break
            cmd += ' %d'% (mem.pop(0) & 0xFFFFFF)
            count_data += 1
            count_argv += 1
            if (len(cmd) >= self.DEFAULT_CMD_LINE_LIMIT) or (count_argv >= self.controller.DEFAULT_CMD_ARGV_LIMIT):
                # split data into multiple lines
                self.controller.writeCommand( cmd )
                if not push:
                    offset += count_data
                cmd = 'W'
                if offset:
                    cmd += ' -o%d'% offset
                if self.swap_rg:
                    cmd += ' -g'
                if push is not None:
                    if push == 'f':
                        cmd += ' -f'
                    elif push == 'b':
                        cmd += ' -b'
                count_data = 0
                count_argv = len(cmd.split())

    def pushf( self, mem, offset=0 ):
        self.write( mem, offset, 'f' )

    def pushb( self, mem, offset=0 ):
        self.write( mem, offset, 'b' )


class LEDS8(LEDS):
    width, height = 8, 1

class LEDS16(LEDS):
    width, height = 16, 1

class LEDS32(LEDS):
    width, height = 32, 1

class LEDS64(LEDS):
    width, height = 64, 1

class LEDS128(LEDS):
    width, height = 128, 1

class LEDS256(LEDS):
    width, height = 256, 1


class LEDS8x8(LEDS):
    width, height = 8, 8

class LEDS16x8(LEDS):
    width, height = 16, 8

class LEDS24x8(LEDS):
    width, height = 24, 8

class LEDS32x8(LEDS):
    width, height = 32, 8
 
class LEDS8x16(LEDS):
    width, height = 8, 16

class LEDS16x16(LEDS):
    width, height = 16, 16

class LEDS24x16(LEDS):
    width, height = 24, 16

class LEDS32x16(LEDS):
    width, height = 32, 16

class LEDS8x24(LEDS):
    width, height = 8, 24

class LEDS16x24(LEDS):
    width, height = 16, 24

class LEDS24x24(LEDS):
    width, height = 24, 24

class LEDS32x24(LEDS):
    width, height = 32, 24

class LEDS8x32(LEDS):
    width, height = 8, 32

class LEDS16x32(LEDS):
    width, height = 16, 32

class LEDS24x32(LEDS):
    width, height = 24, 32

class LEDS32x32(LEDS):
    width, height = 32, 32







# coding:utf8
__doc__ = 'Shell Kit Series'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Mcush


class ShellKit(Mcush.Mcush):
    DEFAULT_NAME = 'ShellKit'
    DEFAULT_IDN = re_compile( 'ShellKit,([0-9]+\.[0-9]+.*)' )

    def scpiRst( self ):
        if self.checkCommand("*rst"):
            self.writeCommand('*rst')
        else:
            self.errno( 0 )
            # clear all leds
            for i in range(self.getLedNumber()):
                self.led(i, False)
            # stop sgpio
            self.sgpioStop()




# coding: utf8
__doc__ = 'products from Dongguan Tongmen Electronics Tech.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
import re
import time
from .. import Env, Instrument
import pymodbus
import pymodbus.client
import pymodbus.client.sync
from pymodbus.client.sync import ModbusSerialClient
from pymodbus.exceptions import ModbusIOException

DEFAULT_UNIT = 0x01  # default instrument address

REG_OUTPUT      = 0x0001
REG_PROTECT     = 0x0002
REG_MODEL       = 0x0003
REG_CLASS       = 0x0004
REG_DIGIT       = 0x0005
REG_VOLTAGE_DISP = 0x0010
REG_CURRENT_DISP = 0x0011
REG_POWER_DISP   = 0x0011
REG_VOLTAGE_SET = 0x0030
REG_CURRENT_SET = 0x0031
REG_OVP_SET = 0x0020
REG_OCP_SET = 0x0021
REG_OPP_SET = 0x0022


class ETM( Instrument.SerialInstrument ):
    DEFAULT_CHECK_RETURN_COMMAND = False
    DEFAULT_IDN = re.compile( '.*' )

    def __init__( self, *args, **kwargs ):
        kwargs['baudrate'] = 9600  # this is fixed
        if not 'unit' in kwargs:
            kwargs['unit'] = DEFAULT_UNIT
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 

    def connect( self, check_idn=True ):
        self.client = ModbusSerialClient('rtu', port=self.port.port, baudrate=self.baudrate, timeout=self.port.timeout)
        self.client.connect()
         
    def outputEnable( self ):
        self.client.write_register(REG_OUTPUT, 1, unit=self.unit)

    def outputDisable( self ):
        self.client.write_register(REG_OUTPUT, 0, unit=self.unit)
 
    def output( self, volt=None, amp=None, enable=True ):
        if volt is not None:
            volt = float(volt)
            if volt < 0 or volt > self.VMAX:
                raise Exception('Volt out of range')
            self.client.write_register(REG_VOLTAGE_SET, int(volt*100), unit=self.unit)
        if amp is not None:
            amp = float(amp)
            if amp < 0 or amp > self.IMAX:
                raise Exception('Current out of range')
            self.client.write_register(REG_CURRENT_SET, int(amp*1000), unit=self.unit)
        if enable is not None:
            if enable:
                self.outputEnable()
            else:
                self.outputDisable()
        
    def getSetting( self ):
        response = self.client.read_holding_registers(REG_VOLTAGE_SET, 1, unit=self.unit)
        if isinstance(response, ModbusIOException):
            raise Instrument.CommandTimeoutError
        regs = response.registers
        volt = regs[0] / 100.0
        response = self.client.read_holding_registers(REG_CURRENT_SET, 1, unit=self.unit)
        if isinstance(response, ModbusIOException):
            raise Instrument.CommandTimeoutError
        regs = response.registers
        amp = regs[0] / 1000.0
        return float(volt), float(amp)

    def readVoltage( self ):
        response = self.client.read_holding_registers(REG_VOLTAGE_DISP, 1, unit=self.unit)
        if isinstance(response, ModbusIOException):
            raise Instrument.CommandTimeoutError
        regs = response.registers
        return regs[0] / 100.0
         
    def readCurrent( self ):
        response = self.client.read_holding_registers(REG_CURRENT_DISP, 1, unit=self.unit)
        if isinstance(response, ModbusIOException):
            raise Instrument.CommandTimeoutError
        regs = response.registers
        return regs[0] / 1000.0

    def readPower( self ):
        response = self.client.read_holding_registers(REG_CURRENT_DISP, 2, unit=self.unit)
        if isinstance(response, ModbusIOException):
            raise Instrument.CommandTimeoutError
        regs = response.registers
        return ((regs[0]<<16)+regs[1]) / 1000.0

    def read( self ):
        return (self.readVoltage(), self.readCurrent())



class ETM605P( ETM ):
    DEFAULT_NAME = 'ETM605P'
    VMAX = 60.0
    IMAX = 5.0




# coding: utf8
# control GRATTEN Instrument with serial port 
# manufactured by Nanjing GLARUN-ATTEN Technology Co. Ltd
# Peng Shulin <trees_peng@163.com> 2018
import os
import re
import sys
import time
from .. import Env, Instrument

WAVEFORM = {
0: 'Sine',
1: "Square",
2: "Triangle",
3: "Up ramp",
4: "Down ramp",
5: "Pos-pulse",
6: "Neg-pulse",
7: "Tri-pulse",
8: "Up-stair",
9: "Pos-DC",
10: "Neg-DC",
11: "All sine",
12: "Half sine",
13: "Limit sine",
14: "Gate sine",
15: "Squar-root",
16: "Exponent",
17: "Logarithm",
18: "Half round",
19: "Tangent",
20: "Sine(x)/x",
21: "Noise",
22: "Duty 10%",
23: "Duty 90%",
24: "Down stair",
25: "Po-bipulse",
26: "Ne-bipulse",
27: "Trapezia",
28: "Cosine",
29: "Bidir-SCR",
30: "Cardiogram",
31: "Earthquake",
}

WAVEFORM_REV = {}
for k, v in WAVEFORM.items():
    WAVEFORM_REV[v] = k
    WAVEFORM_REV[v.lower()] = k
    WAVEFORM_REV[v.upper()] = k


def convertUnit( value ):
    leading = ''
    value_abs = abs(value)
    if value_abs > 1000000.0:
        value, leading = value/1000000.0, 'M'
    elif value_abs > 1000.0:
        value, leading = value/1000.0, 'k'
    elif value_abs < 0.001:
        value, leading = value*1000.0, 'm'
    elif value_abs < 0.000001:
        value, leading = value*1000000.0, 'u'
    return value, leading
    
    
def getWaveformIndex( name ):
    if isinstance( name, int ):
        return name
    elif isinstance( name, str ):
        return WAVEFORM_REV[name]


class ATF20E( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'ATF20E'

    def __init__( self, *args, **kwargs ):
        if not 'address' in kwargs:
            kwargs['address'] = Env.getenv_int('ADDRESS', 88) 
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 
        
    def connect( self, check_idn=False ):
        '''connect'''
        self.port.parity = 'M'
        self.port.timeout = 1
        self.port.connect()
        if not self.port.connected:
            return
        self.command( 'rs232' )

    def disconnect( self ):
        self.command( 'local' )
 
    def writeCommand( self, cmd ):
        self.port.write( chr(self.address) )
        self.port.flush()
        self.port.disconnect()
        self.port.parity = 'S'
        self.port.connect()
        self.logger.debug( cmd )
        self.port.write(cmd + '\n')
        self.port.flush()
        ret = self.readLine()
        ret_addr = ord(ret[0])
        assert ret_addr == self.address
        ret_contents = ret[1:]
        self.logger.debug( ret_contents )
        if ret_contents == 'Receive error':
            raise Instrument.CommandSyntaxError()
        #time.sleep(5)
        return ret_contents

    def readLine( self ):
        newline_str = ''
        while True:
            byte = self.port.read(1)
            if byte:
                if Env.PYTHON_V3:
                    byte = chr(ord(byte))
                if byte == self.DEFAULT_TERMINATOR_READ:
                    return newline_str
                else:
                    newline_str += byte
            else:
                raise Instrument.CommandTimeoutError( 'No respone: ' + newline_str )

    def command( self, function, option=None, data=None, unit=None ):
        cmd = str(function).upper()
        if option is not None:
            cmd += ':%s'% option.upper()
            if data is not None:
                cmd += (':%s'% data).upper()
                if unit is not None:
                    cmd += ':%s'% unit
        return self.writeCommand( cmd ) 

    def reset(self):
        try:
            self.command( 'sys', 'reset' )
        except Instrument.CommandTimeoutError as e:
            pass
        time.sleep(5)
        while 1:
            try:
                self.command( 'rs232' )
                break
            except Instrument.CommandSyntaxError as e:
                pass
            
            
    def getVersion(self):
        return self.command( 'sys', '?ver' )

    def disableA( self ):
        self.command( 'cha', 'switch', 0, 'No.' )

    def disableB( self ):
        self.command( 'chb', 'switch', 0, 'No.' )

    def singleA( self, mode, freq, vpp, offset=0 ):
        switch = self.command( 'cha', '?switch' ).split()[-1]
        if switch == 'ON':
            self.command( 'cha', 'switch', 0, 'No.' )
        self.command( 'cha', 'awave', getWaveformIndex(mode), 'No.' )
        val, lu = convertUnit(vpp)
        self.command( 'cha', 'vpp', val, lu+'Vpp' )
        val, lu = convertUnit(freq)
        self.command( 'cha', 'afreq', val, lu+'Hz' )
        val, lu = convertUnit(offset)
        self.command( 'cha', 'aoffs', val, lu+'Vdc' )
        self.command( 'cha', 'switch', 1, 'No.' )
        #print self.command( 'cha', '?aoffs' )
        #print self.command( 'cha', '?aatte' )
        #print self.command( 'cha', '?afreq' )
    
    def singleB( self, mode, freq, vpp ):
        switch = self.command( 'chb', '?switch' ).split()[-1]
        if switch == 'ON':
            self.command( 'chb', 'switch', 0, 'No.' )
        self.command( 'chb', 'bwave', getWaveformIndex(mode), 'No.' )
        val, lu = convertUnit(freq)
        self.command( 'chb', 'bfreq', val, lu+'Hz' )
        val, lu = convertUnit(vpp)
        self.command( 'chb', 'vpp', val, lu+'Vpp' )
        self.command( 'chb', 'switch', 1, 'No.' )







# coding: utf8
__doc__ = 'Bosch sensors'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import time


REG_AC1 = 0xAA
REG_AC2 = 0xAC
REG_AC3 = 0xAE
REG_AC4 = 0xB0
REG_AC5 = 0xB2
REG_AC6 = 0xB4
REG_B1 = 0xB6
REG_B2 = 0xB8
REG_MB = 0xBA
REG_MC = 0xBC
REG_MD = 0xBE

REG_CTRL_MEAS = 0xF4
REG_OUT = 0xF6
REG_OUT_XLSB = 0xF8
REG_SOFT_RESET = 0xE0
REG_ID = 0xD0


class Bmp180():
        
    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x77, scl, sda )
        self.reset()
        self.init()

    def readReg( self, addr, signed=False ):
        h, l = self.controller.i2c( [addr], read=2 )
        r = (h<<8) + l
        return Utils.H2h(r) if signed else r

    def readReg3( self, addr ):
        h, l, x = self.controller.i2c( [addr], read=3 )
        r = (h<<16) + (l<<8) + x
        return r

    def init( self ):
        # read calibration parameters
        self.ac1 = self.readReg( REG_AC1, True )
        self.ac2 = self.readReg( REG_AC2, True )
        self.ac3 = self.readReg( REG_AC3, True )
        self.ac4 = self.readReg( REG_AC4 )
        self.ac5 = self.readReg( REG_AC5 )
        self.ac6 = self.readReg( REG_AC6 )
        self.b1 = self.readReg( REG_B1, True )
        self.b2 = self.readReg( REG_B2, True )
        self.mb = self.readReg( REG_MB, True )
        self.mc = self.readReg( REG_MC, True )
        self.md = self.readReg( REG_MD, True )
        self.printCalibRegs()

    def printCalibRegs( self ):
        print( 'AC1: %d'% (self.ac1) )
        print( 'AC2: %d'% (self.ac2) )
        print( 'AC3: %d'% (self.ac3) )
        print( 'AC4: %d'% (self.ac4) )
        print( 'AC5: %d'% (self.ac5) )
        print( 'AC6: %d'% (self.ac6) )
        print( 'B1:  %d'% (self.b1) )
        print( 'B2:  %d'% (self.b2) )
        print( 'MB:  %d'% (self.mb) )
        print( 'MC:  %d'% (self.mc) )
        print( 'MD:  %d'% (self.md) )

    def reset( self ):
        self.controller.i2c( [REG_SOFT_RESET, 0xB6] )
        time.sleep(0.1)

    def read( self, oss=3 ):
        # read temperature
        self.controller.i2c( [REG_CTRL_MEAS, 0x2E] )
        time.sleep(0.005)
        ut = self.readReg( REG_OUT )
        # read pressure
        mode = 0x34 | (oss<<6)
        self.controller.i2c( [REG_CTRL_MEAS, mode] )
        time.sleep(0.005+0.003*(2**oss))
        up = self.readReg3( REG_OUT ) >> (8-oss)
        # calculate temperature
        x1 = ((ut - self.ac6) * self.ac5) / (2**15)
        x2 = (self.mc * (2**11) ) / (x1 + self.md)
        b5 = x1 + x2
        t = (b5 + 8) / (2**4)
        # calculate pressure
        b6 = b5 - 4000
        x1 = (self.b2 * (b6 * b6 / (2**12)))/(2**11)
        x2 = self.ac2 * b6 / (2**11)
        x3 = x1 + x2
        b3 = (((self.ac1*4 + x3)<<oss) + 2) / 4
        x1 = self.ac3 * b6 / (2**13)
        x2 = (self.b1 * (b6 * b6 / (2**12))) / (2**16)
        x3 = ((x1 + x2) + 2) / (2**2)
        b4 = self.ac4 * (x3 + 32768) / (2**15)
        b7 = (up - b3) * (50000>>oss)
        if b7 < 0x80000000:
            p = (b7 * 2) / b4
        else:
            p = (b7 / b4) * 2 
        x1 = (p / (2**8))**2
        x1 = (x1 * 3038) / (2**16)
        x2 = (-7357 * p) / (2**16)
        p = p + (x1 + x2 + 3791) / (2**4)
        return float(p), float(t)/10

    def convertPressure2Altitude( self, p ):
        a = 44330 * (1-((p/101325)**(1/5.255)))
        return a




# coding: utf8
__doc__ = 'CS1237/1238 24-bit ADC'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import time


CMD_WRITE_CONFIG = 0x65
CMD_READ_CONFIG = 0x56

# config
REF_OUT_ON  = 0<<6
REF_OUT_OFF = 1<<6
SPEED_10    = 0<<4
SPEED_40    = 1<<4
SPEED_640   = 2<<4
SPEED_1280  = 3<<4
GAIN_1      = 0<<2
GAIN_2      = 1<<2
GAIN_64     = 2<<2
GAIN_128    = 3<<2
CH_A        = 0
CH_B        = 1  # CS1238 only
CH_T        = 2
CH_SHORT    = 3


class Cs1238():
    ref = 3.3
    channel = CH_A
    gain = GAIN_1
    speed = SPEED_10
    ref_out = REF_OUT_ON

    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None, channel='a', gain=1, speed=10, ref_out=True, ref=3.3, init=True ):
        self.controller = controller
        self.sdi = sdi if sdi is not None else '0.0'
        self.sdo = sdo if sdo is not None else '0.1'
        self.sck = sck if sck is not None else '0.2'
        self.cs  = cs  if cs  is not None else '0.3'
        self.controller.spiInit(self.sdi, self.sdo, self.sck, self.cs, width=27, cpha=True)
        if channel in [0,'a','A']:
            self.channel = CH_A
        elif channel in [1,'b','B']:
            self.channel = CH_B
        elif channel in [2,'t','T']:
            self.channel = CH_T
        elif channel in [3,'s','S']:
            self.channel = CH_SHORT
        if gain == 1:
            self.gain = GAIN_1
        elif gain == 2:
            self.gain = GAIN_2
        elif gain == 64:
            self.gain = GAIN_64
        elif gain == 128:
            self.gain = GAIN_128
        if speed == 10:
            self.speed = SPEED_10
        elif speed == 40:
            self.speed = SPEED_40
        elif speed == 640:
            self.speed = SPEED_640
        elif speed == 1280:
            self.speed = SPEED_1280
        self.ref_out = REF_OUT_ON if ref_out else REF_OUT_OFF
        self.ref = ref
        if init:
            self.writeConfig()

    def writeConfig( self, read=False ):
        # wait for drdy
        while not self.dataIsReady():
            time.sleep(0.1)
        # CLK 1~27: read adc
        # CLK 28~29: read 2 dummy clocks
        self.controller.spiUpdate(self.sdi, self.sdo, self.sck, self.cs, width=27+2, cpha=True)
        self.controller.spiRead( [0] )
        # CLK 30~36: output command
        self.controller.spiUpdate(self.sdi, self.sdi, self.sck, self.cs, width=7, cpha=True)
        self.controller.spiWrite( [CMD_READ_CONFIG if read else CMD_WRITE_CONFIG] )
        # CLK 37: switch sdi as input for read mode
        # CLK 38~45: read/write config
        if read:
            self.controller.spiUpdate(self.sdi, self.sdo, self.sck, self.cs, width=9, cpha=True)
            ret = self.controller.spiRead([0])[0] & 0xFF
        else:
            self.controller.spiUpdate(self.sdi, self.sdi, self.sck, self.cs, width=9, cpha=True)
            self.controller.spiWrite( [self.ref_out | self.speed | self.gain | self.channel] )
        # CLK 46: switch back
        self.controller.spiUpdate(self.sdi, self.sdo, self.sck, self.cs, width=1, cpha=True)
        self.controller.spiWrite( [0] )
        self.controller.spiUpdate(self.sdi, self.sdo, self.sck, self.cs, width=27, cpha=True)
        if read:
            return ret

    def readConfig( self ):
        return self.writeConfig(True)

    def dataIsReady( self ):
        return self.controller.pinIsLow(self.sdi)

    def hex2volt( self, dat ):
        val = Utils.I2i(dat<<8)/256
        return self.ref * val / (2**24)
        
    def readData( self, mode=None ):
        a = self.controller.spiRead([0])[0]
        if mode == 'hex':
            return a>>3
        else:
            val = Utils.I2i(a<<5)/256
            if mode == 'int':
                return val
            return self.ref * val / (2**24)


